"""SQL Dialect-specific Code Generators.

This module provides dialect-specific SQL generators that convert
AST nodes into valid SQL strings for each database platform.

Supported Dialects:
    - PostgreSQL
    - MySQL
    - SQLite
    - BigQuery
    - Snowflake
    - Redshift
    - Databricks (Spark SQL)
    - Oracle
    - SQL Server (T-SQL)

Example:
    >>> from truthound.execution.pushdown.dialects import (
    ...     SQLDialect,
    ...     get_dialect_generator,
    ... )
    >>> from truthound.execution.pushdown.ast import SelectStatement, Column, Table
    >>>
    >>> stmt = SelectStatement(...)
    >>> generator = get_dialect_generator(SQLDialect.POSTGRESQL)
    >>> sql = generator.generate(stmt)
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, Sequence

from truthound.execution.pushdown.ast import (
    # Base
    SQLNode,
    SQLVisitor,
    Expression,
    # Literals
    Literal,
    NullLiteral,
    BooleanLiteral,
    ArrayLiteral,
    # Identifiers
    Identifier,
    Column,
    Table,
    Alias,
    Star,
    # Operators
    ComparisonOp,
    LogicalOp,
    ArithmeticOp,
    BinaryOp,
    UnaryOp,
    JoinType,
    SortOrder,
    NullsPosition,
    FrameType,
    FrameBoundType,
    SetOperation,
    # Expressions
    BinaryExpression,
    UnaryExpression,
    InExpression,
    BetweenExpression,
    ExistsExpression,
    SubqueryExpression,
    CastExpression,
    WhenClause,
    CaseExpression,
    FunctionCall,
    AggregateFunction,
    FrameBound,
    WindowSpec,
    WindowFunction,
    # Clauses
    SelectItem,
    FromClause,
    JoinClause,
    WhereClause,
    GroupByClause,
    HavingClause,
    OrderByItem,
    OrderByClause,
    LimitClause,
    OffsetClause,
    CTEClause,
    # Statements
    SelectStatement,
    SetOperationStatement,
)


# =============================================================================
# SQL Dialect Enum
# =============================================================================


class SQLDialect(Enum):
    """Supported SQL dialects."""

    POSTGRESQL = "postgresql"
    MYSQL = "mysql"
    SQLITE = "sqlite"
    BIGQUERY = "bigquery"
    SNOWFLAKE = "snowflake"
    REDSHIFT = "redshift"
    DATABRICKS = "databricks"
    ORACLE = "oracle"
    SQLSERVER = "sqlserver"
    DUCKDB = "duckdb"
    CLICKHOUSE = "clickhouse"
    PRESTO = "presto"
    TRINO = "trino"
    HIVE = "hive"
    GENERIC = "generic"


# =============================================================================
# Dialect Configuration
# =============================================================================


@dataclass
class DialectConfig:
    """Configuration for SQL dialect.

    Attributes:
        identifier_quote: Character(s) to quote identifiers.
        string_quote: Character to quote strings.
        escape_char: Character for escaping.
        supports_nulls_ordering: Whether NULLS FIRST/LAST is supported.
        supports_limit_offset: Whether LIMIT/OFFSET syntax is supported.
        supports_fetch_first: Whether FETCH FIRST syntax is supported.
        supports_top: Whether TOP syntax is supported (SQL Server).
        supports_window_functions: Whether window functions are supported.
        supports_cte: Whether WITH clause (CTEs) are supported.
        supports_lateral: Whether LATERAL joins are supported.
        supports_json: Whether JSON functions are supported.
        supports_arrays: Whether array types are supported.
        supports_filter_clause: Whether FILTER (WHERE ...) clause is supported.
        supports_grouping_sets: Whether GROUPING SETS are supported.
        max_identifier_length: Maximum identifier length.
        case_sensitive_identifiers: Whether identifiers are case-sensitive.
        boolean_as_int: Whether booleans are represented as integers.
        concat_operator: Operator for string concatenation.
        date_format: Default date format string.
        timestamp_format: Default timestamp format string.
    """

    identifier_quote: str = '"'
    string_quote: str = "'"
    escape_char: str = "\\"
    supports_nulls_ordering: bool = True
    supports_limit_offset: bool = True
    supports_fetch_first: bool = False
    supports_top: bool = False
    supports_window_functions: bool = True
    supports_cte: bool = True
    supports_lateral: bool = False
    supports_json: bool = True
    supports_arrays: bool = False
    supports_filter_clause: bool = False
    supports_grouping_sets: bool = True
    max_identifier_length: int = 128
    case_sensitive_identifiers: bool = False
    boolean_as_int: bool = False
    concat_operator: str = "||"
    date_format: str = "YYYY-MM-DD"
    timestamp_format: str = "YYYY-MM-DD HH24:MI:SS"


# Predefined configs for common dialects
DIALECT_CONFIGS: dict[SQLDialect, DialectConfig] = {
    SQLDialect.POSTGRESQL: DialectConfig(
        identifier_quote='"',
        supports_nulls_ordering=True,
        supports_filter_clause=True,
        supports_lateral=True,
        supports_arrays=True,
    ),
    SQLDialect.MYSQL: DialectConfig(
        identifier_quote="`",
        supports_nulls_ordering=False,
        supports_filter_clause=False,
        concat_operator="CONCAT",
    ),
    SQLDialect.SQLITE: DialectConfig(
        identifier_quote='"',
        supports_nulls_ordering=False,
        supports_cte=True,
        supports_window_functions=True,
        supports_lateral=False,
    ),
    SQLDialect.BIGQUERY: DialectConfig(
        identifier_quote="`",
        supports_nulls_ordering=True,
        supports_arrays=True,
        supports_lateral=True,
    ),
    SQLDialect.SNOWFLAKE: DialectConfig(
        identifier_quote='"',
        supports_nulls_ordering=True,
        supports_lateral=True,
        case_sensitive_identifiers=False,
    ),
    SQLDialect.REDSHIFT: DialectConfig(
        identifier_quote='"',
        supports_nulls_ordering=True,
        supports_lateral=False,
        max_identifier_length=127,
    ),
    SQLDialect.DATABRICKS: DialectConfig(
        identifier_quote="`",
        supports_nulls_ordering=True,
        supports_arrays=True,
        supports_lateral=True,
    ),
    SQLDialect.ORACLE: DialectConfig(
        identifier_quote='"',
        supports_limit_offset=False,
        supports_fetch_first=True,
        supports_nulls_ordering=True,
        max_identifier_length=30,
    ),
    SQLDialect.SQLSERVER: DialectConfig(
        identifier_quote="[",
        supports_limit_offset=False,
        supports_top=True,
        supports_nulls_ordering=False,
        concat_operator="+",
    ),
    SQLDialect.DUCKDB: DialectConfig(
        identifier_quote='"',
        supports_nulls_ordering=True,
        supports_filter_clause=True,
        supports_arrays=True,
    ),
    SQLDialect.GENERIC: DialectConfig(),
}


# =============================================================================
# Base Dialect Generator
# =============================================================================


class BaseDialectGenerator(SQLVisitor, ABC):
    """Base class for SQL dialect generators.

    This class implements the visitor pattern to convert AST nodes
    into SQL strings. Subclasses can override methods to customize
    SQL generation for specific dialects.

    Attributes:
        config: Dialect configuration.
        indent_level: Current indentation level.
        indent_str: String used for indentation.
    """

    dialect: SQLDialect = SQLDialect.GENERIC

    def __init__(
        self,
        config: DialectConfig | None = None,
        indent_str: str = "  ",
    ) -> None:
        """Initialize generator.

        Args:
            config: Optional dialect configuration.
            indent_str: String used for indentation.
        """
        self.config = config or DIALECT_CONFIGS.get(
            self.dialect, DialectConfig()
        )
        self.indent_level = 0
        self.indent_str = indent_str
        self._function_mappings: dict[str, str] = {}

    def generate(self, node: SQLNode) -> str:
        """Generate SQL string from AST node.

        Args:
            node: AST node to generate SQL for.

        Returns:
            SQL string.
        """
        return node.accept(self)

    def _indent(self) -> str:
        """Get current indentation string."""
        return self.indent_str * self.indent_level

    def _quote_identifier(self, name: str) -> str:
        """Quote an identifier.

        Args:
            name: Identifier name.

        Returns:
            Quoted identifier.
        """
        quote = self.config.identifier_quote
        if quote == "[":
            return f"[{name}]"
        return f"{quote}{name}{quote}"

    def _quote_string(self, value: str) -> str:
        """Quote a string literal.

        Args:
            value: String value.

        Returns:
            Quoted string literal.
        """
        quote = self.config.string_quote
        escaped = value.replace(quote, quote + quote)
        return f"{quote}{escaped}{quote}"

    def _format_literal(self, value: Any) -> str:
        """Format a literal value.

        Args:
            value: Value to format.

        Returns:
            SQL literal string.
        """
        if value is None:
            return "NULL"
        if isinstance(value, bool):
            if self.config.boolean_as_int:
                return "1" if value else "0"
            return "TRUE" if value else "FALSE"
        if isinstance(value, str):
            return self._quote_string(value)
        if isinstance(value, (int, float)):
            return str(value)
        if isinstance(value, bytes):
            return f"X'{value.hex()}'"
        return self._quote_string(str(value))

    def _map_function(self, name: str) -> str:
        """Map a function name to dialect-specific name.

        Args:
            name: Function name.

        Returns:
            Dialect-specific function name.
        """
        return self._function_mappings.get(name.upper(), name)

    # -------------------------------------------------------------------------
    # Visitor Methods - Literals
    # -------------------------------------------------------------------------

    def visit_literal(self, node: Literal) -> str:
        return self._format_literal(node.value)

    def visit_null_literal(self, node: NullLiteral) -> str:
        return "NULL"

    def visit_boolean_literal(self, node: BooleanLiteral) -> str:
        if self.config.boolean_as_int:
            return "1" if node.value else "0"
        return "TRUE" if node.value else "FALSE"

    def visit_array_literal(self, node: ArrayLiteral) -> str:
        elements = ", ".join(e.accept(self) for e in node.elements)
        return f"ARRAY[{elements}]"

    # -------------------------------------------------------------------------
    # Visitor Methods - Identifiers
    # -------------------------------------------------------------------------

    def visit_identifier(self, node: Identifier) -> str:
        if node.quoted:
            return self._quote_identifier(node.name)
        return node.name

    def visit_column(self, node: Column) -> str:
        parts = []
        if node.schema:
            parts.append(self._quote_identifier(node.schema))
        if node.table:
            parts.append(self._quote_identifier(node.table))

        if node.name == "*":
            parts.append("*")
        else:
            parts.append(self._quote_identifier(node.name))

        return ".".join(parts)

    def visit_table(self, node: Table) -> str:
        parts = []
        if node.catalog:
            parts.append(self._quote_identifier(node.catalog))
        if node.schema:
            parts.append(self._quote_identifier(node.schema))
        parts.append(self._quote_identifier(node.name))

        result = ".".join(parts)
        if node.alias:
            result += f" AS {self._quote_identifier(node.alias)}"
        return result

    def visit_alias(self, node: Alias) -> str:
        expr = node.expression.accept(self)
        return f"{expr} AS {self._quote_identifier(node.alias)}"

    def visit_star(self, node: Star) -> str:
        if node.table:
            return f"{self._quote_identifier(node.table)}.*"
        return "*"

    # -------------------------------------------------------------------------
    # Visitor Methods - Expressions
    # -------------------------------------------------------------------------

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        left = node.left.accept(self)
        right = node.right.accept(self)

        # Get operator string
        if isinstance(node.operator, Enum):
            op = node.operator.value
        else:
            op = str(node.operator)

        # Handle special cases
        if op == "||" and self.config.concat_operator == "CONCAT":
            return f"CONCAT({left}, {right})"

        return f"({left} {op} {right})"

    def visit_unary_expression(self, node: UnaryExpression) -> str:
        operand = node.operand.accept(self)

        if node.operator in (UnaryOp.IS_NULL, UnaryOp.IS_NOT_NULL):
            return f"({operand} {node.operator.value})"

        if node.operator in (UnaryOp.EXISTS, UnaryOp.NOT_EXISTS):
            return f"{node.operator.value} ({operand})"

        return f"({node.operator.value} {operand})"

    def visit_in_expression(self, node: InExpression) -> str:
        expr = node.expression.accept(self)
        op = "NOT IN" if node.negated else "IN"

        if isinstance(node.values, SelectStatement):
            subquery = node.values.accept(self)
            return f"{expr} {op} ({subquery})"
        else:
            values = ", ".join(v.accept(self) for v in node.values)
            return f"{expr} {op} ({values})"

    def visit_between_expression(self, node: BetweenExpression) -> str:
        expr = node.expression.accept(self)
        low = node.low.accept(self)
        high = node.high.accept(self)
        op = "NOT BETWEEN" if node.negated else "BETWEEN"
        return f"{expr} {op} {low} AND {high}"

    def visit_exists_expression(self, node: ExistsExpression) -> str:
        subquery = node.subquery.accept(self)
        op = "NOT EXISTS" if node.negated else "EXISTS"
        return f"{op} ({subquery})"

    def visit_subquery_expression(self, node: SubqueryExpression) -> str:
        subquery = node.subquery.accept(self)
        return f"({subquery})"

    def visit_cast_expression(self, node: CastExpression) -> str:
        expr = node.expression.accept(self)
        return f"CAST({expr} AS {node.target_type})"

    def visit_when_clause(self, node: WhenClause) -> str:
        condition = node.condition.accept(self)
        result = node.result.accept(self)
        return f"WHEN {condition} THEN {result}"

    def visit_case_expression(self, node: CaseExpression) -> str:
        parts = ["CASE"]

        if node.operand:
            parts.append(node.operand.accept(self))

        for when in node.when_clauses:
            parts.append(when.accept(self))

        if node.else_result:
            parts.append(f"ELSE {node.else_result.accept(self)}")

        parts.append("END")
        return " ".join(parts)

    # -------------------------------------------------------------------------
    # Visitor Methods - Functions
    # -------------------------------------------------------------------------

    def visit_function_call(self, node: FunctionCall) -> str:
        name = self._map_function(node.name)
        distinct = "DISTINCT " if node.distinct else ""

        if node.arguments:
            args = ", ".join(a.accept(self) for a in node.arguments)
        else:
            args = ""

        result = f"{name}({distinct}{args})"

        if node.filter_clause and self.config.supports_filter_clause:
            filter_expr = node.filter_clause.accept(self)
            result += f" FILTER (WHERE {filter_expr})"

        return result

    def visit_aggregate_function(self, node: AggregateFunction) -> str:
        name = self._map_function(node.name)
        distinct = "DISTINCT " if node.distinct else ""

        if node.argument:
            arg = node.argument.accept(self)
        else:
            arg = "*"

        # Handle ordered-set aggregates
        if node.order_by:
            order_items = ", ".join(o.accept(self) for o in node.order_by)
            result = f"{name}({distinct}{arg} ORDER BY {order_items})"
        else:
            result = f"{name}({distinct}{arg})"

        # Add FILTER clause
        if node.filter_clause and self.config.supports_filter_clause:
            filter_expr = node.filter_clause.accept(self)
            result += f" FILTER (WHERE {filter_expr})"

        return result

    def visit_frame_bound(self, node: FrameBound) -> str:
        if node.offset is not None:
            if node.bound_type == FrameBoundType.PRECEDING:
                return f"{node.offset} PRECEDING"
            elif node.bound_type == FrameBoundType.FOLLOWING:
                return f"{node.offset} FOLLOWING"
        return node.bound_type.value

    def visit_window_spec(self, node: WindowSpec) -> str:
        parts = []

        # PARTITION BY
        if node.partition_by:
            partition = ", ".join(e.accept(self) for e in node.partition_by)
            parts.append(f"PARTITION BY {partition}")

        # ORDER BY
        if node.order_by:
            order = ", ".join(o.accept(self) for o in node.order_by)
            parts.append(f"ORDER BY {order}")

        # Frame specification
        if node.frame_type:
            frame_parts = [node.frame_type.value]

            if node.frame_start and node.frame_end:
                start = node.frame_start.accept(self)
                end = node.frame_end.accept(self)
                frame_parts.append(f"BETWEEN {start} AND {end}")
            elif node.frame_start:
                frame_parts.append(node.frame_start.accept(self))

            parts.append(" ".join(frame_parts))

        return " ".join(parts)

    def visit_window_function(self, node: WindowFunction) -> str:
        func = node.function.accept(self)

        if isinstance(node.window_spec, str):
            return f"{func} OVER {node.window_spec}"
        elif node.window_spec:
            spec = node.window_spec.accept(self)
            return f"{func} OVER ({spec})"
        else:
            return f"{func} OVER ()"

    # -------------------------------------------------------------------------
    # Visitor Methods - Clauses
    # -------------------------------------------------------------------------

    def visit_select_item(self, node: SelectItem) -> str:
        expr = node.expression.accept(self)
        if node.alias:
            return f"{expr} AS {self._quote_identifier(node.alias)}"
        return expr

    def visit_from_clause(self, node: FromClause) -> str:
        return node.source.accept(self)

    def visit_join_clause(self, node: JoinClause) -> str:
        left = node.left.accept(self)
        right = node.right.accept(self)

        result = f"{left} {node.join_type.value} {right}"

        if node.condition:
            condition = node.condition.accept(self)
            result += f" ON {condition}"
        elif node.using_columns:
            cols = ", ".join(self._quote_identifier(c) for c in node.using_columns)
            result += f" USING ({cols})"

        return result

    def visit_where_clause(self, node: WhereClause) -> str:
        return node.condition.accept(self)

    def visit_group_by_clause(self, node: GroupByClause) -> str:
        exprs = ", ".join(e.accept(self) for e in node.expressions)

        if node.grouping_sets and self.config.supports_grouping_sets:
            sets = []
            for gs in node.grouping_sets:
                gs_exprs = ", ".join(e.accept(self) for e in gs)
                sets.append(f"({gs_exprs})")
            return f"GROUPING SETS ({', '.join(sets)})"

        result = exprs
        if node.with_rollup:
            result += " WITH ROLLUP"
        elif node.with_cube:
            result += " WITH CUBE"

        return result

    def visit_having_clause(self, node: HavingClause) -> str:
        return node.condition.accept(self)

    def visit_order_by_item(self, node: OrderByItem) -> str:
        expr = node.expression.accept(self)
        parts = [expr, node.order.value]

        if node.nulls and self.config.supports_nulls_ordering:
            parts.append(node.nulls.value)

        return " ".join(parts)

    def visit_order_by_clause(self, node: OrderByClause) -> str:
        return ", ".join(item.accept(self) for item in node.items)

    def visit_limit_clause(self, node: LimitClause) -> str:
        if isinstance(node.count, Expression):
            return node.count.accept(self)
        return str(node.count)

    def visit_offset_clause(self, node: OffsetClause) -> str:
        if isinstance(node.offset, Expression):
            return node.offset.accept(self)
        return str(node.offset)

    def visit_cte_clause(self, node: CTEClause) -> str:
        name = self._quote_identifier(node.name)

        if node.columns:
            cols = ", ".join(self._quote_identifier(c) for c in node.columns)
            name = f"{name}({cols})"

        query = node.query.accept(self)
        return f"{name} AS ({query})"

    # -------------------------------------------------------------------------
    # Visitor Methods - Statements
    # -------------------------------------------------------------------------

    def visit_select_statement(self, node: SelectStatement) -> str:
        parts = []

        # WITH clause (CTEs)
        if node.ctes:
            has_recursive = any(cte.recursive for cte in node.ctes)
            cte_keyword = "WITH RECURSIVE" if has_recursive else "WITH"
            cte_parts = [cte.accept(self) for cte in node.ctes]
            parts.append(f"{cte_keyword} {', '.join(cte_parts)}")

        # SELECT
        distinct = "DISTINCT " if node.distinct else ""
        select_items = ", ".join(
            item.accept(self) if isinstance(item, SelectItem) else item.accept(self)
            for item in node.select_items
        )

        # Handle TOP for SQL Server
        if self.config.supports_top and node.limit_clause and not node.offset_clause:
            limit = self.visit_limit_clause(node.limit_clause)
            parts.append(f"SELECT {distinct}TOP {limit} {select_items}")
        else:
            parts.append(f"SELECT {distinct}{select_items}")

        # FROM
        if node.from_clause:
            parts.append(f"FROM {node.from_clause.accept(self)}")

        # WHERE
        if node.where_clause:
            parts.append(f"WHERE {node.where_clause.accept(self)}")

        # GROUP BY
        if node.group_by_clause:
            parts.append(f"GROUP BY {node.group_by_clause.accept(self)}")

        # HAVING
        if node.having_clause:
            parts.append(f"HAVING {node.having_clause.accept(self)}")

        # ORDER BY
        if node.order_by_clause:
            parts.append(f"ORDER BY {node.order_by_clause.accept(self)}")

        # LIMIT/OFFSET or FETCH FIRST
        if node.limit_clause and not self.config.supports_top:
            if self.config.supports_fetch_first:
                limit = self.visit_limit_clause(node.limit_clause)
                if node.offset_clause:
                    offset = self.visit_offset_clause(node.offset_clause)
                    parts.append(f"OFFSET {offset} ROWS FETCH FIRST {limit} ROWS ONLY")
                else:
                    parts.append(f"FETCH FIRST {limit} ROWS ONLY")
            elif self.config.supports_limit_offset:
                limit = self.visit_limit_clause(node.limit_clause)
                parts.append(f"LIMIT {limit}")
                if node.offset_clause:
                    offset = self.visit_offset_clause(node.offset_clause)
                    parts.append(f"OFFSET {offset}")

        return "\n".join(parts)

    def visit_set_operation(self, node: SetOperationStatement) -> str:
        left = node.left.accept(self)
        right = node.right.accept(self)
        return f"({left})\n{node.operation.value}\n({right})"


# =============================================================================
# Dialect-Specific Generators
# =============================================================================


class PostgreSQLGenerator(BaseDialectGenerator):
    """SQL generator for PostgreSQL."""

    dialect = SQLDialect.POSTGRESQL

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "LENGTH": "LENGTH",
            "SUBSTR": "SUBSTRING",
            "NVL": "COALESCE",
            "IFNULL": "COALESCE",
            "NOW": "NOW",
            "GETDATE": "NOW",
        }

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        # Handle ILIKE (PostgreSQL native)
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.ILIKE:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"({left} ILIKE {right})"

        # Handle REGEXP
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.REGEXP:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"({left} ~ {right})"

        return super().visit_binary_expression(node)


class MySQLGenerator(BaseDialectGenerator):
    """SQL generator for MySQL."""

    dialect = SQLDialect.MYSQL

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "LENGTH": "CHAR_LENGTH",
            "NVL": "IFNULL",
            "COALESCE": "COALESCE",
            "NOW": "NOW",
            "CURRENT_TIMESTAMP": "NOW",
        }

    def visit_boolean_literal(self, node: BooleanLiteral) -> str:
        # MySQL uses TRUE/FALSE
        return "TRUE" if node.value else "FALSE"

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        # Handle ILIKE (convert to case-insensitive LIKE)
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.ILIKE:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"(LOWER({left}) LIKE LOWER({right}))"

        # Handle REGEXP
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.REGEXP:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"({left} REGEXP {right})"

        return super().visit_binary_expression(node)


class SQLiteGenerator(BaseDialectGenerator):
    """SQL generator for SQLite."""

    dialect = SQLDialect.SQLITE

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "NVL": "IFNULL",
            "COALESCE": "COALESCE",
            "NOW": "DATETIME",
            "LEN": "LENGTH",
        }

    def visit_boolean_literal(self, node: BooleanLiteral) -> str:
        # SQLite uses 1/0
        return "1" if node.value else "0"


class BigQueryGenerator(BaseDialectGenerator):
    """SQL generator for Google BigQuery."""

    dialect = SQLDialect.BIGQUERY

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "NVL": "IFNULL",
            "LENGTH": "LENGTH",
            "SUBSTR": "SUBSTR",
            "NOW": "CURRENT_TIMESTAMP",
        }

    def visit_array_literal(self, node: ArrayLiteral) -> str:
        elements = ", ".join(e.accept(self) for e in node.elements)
        return f"[{elements}]"

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        # Handle REGEXP
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.REGEXP:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"REGEXP_CONTAINS({left}, {right})"

        return super().visit_binary_expression(node)


class SnowflakeGenerator(BaseDialectGenerator):
    """SQL generator for Snowflake."""

    dialect = SQLDialect.SNOWFLAKE

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "LENGTH": "LENGTH",
            "SUBSTR": "SUBSTR",
            "NVL": "NVL",
            "NOW": "CURRENT_TIMESTAMP",
        }

    def _quote_identifier(self, name: str) -> str:
        # Snowflake preserves case with quotes
        quote = self.config.identifier_quote
        return f'{quote}{name}{quote}'

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        # Handle REGEXP
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.REGEXP:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"REGEXP_LIKE({left}, {right})"

        # Handle ILIKE (Snowflake native)
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.ILIKE:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"({left} ILIKE {right})"

        return super().visit_binary_expression(node)


class RedshiftGenerator(BaseDialectGenerator):
    """SQL generator for Amazon Redshift."""

    dialect = SQLDialect.REDSHIFT

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "LENGTH": "LEN",
            "SUBSTR": "SUBSTRING",
            "NVL": "NVL",
            "NOW": "GETDATE",
        }

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        # Handle REGEXP (Redshift uses SIMILAR TO or REGEXP_SUBSTR)
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.REGEXP:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"({left} ~ {right})"

        # Handle ILIKE (Redshift native)
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.ILIKE:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"({left} ILIKE {right})"

        return super().visit_binary_expression(node)


class DatabricksGenerator(BaseDialectGenerator):
    """SQL generator for Databricks (Spark SQL)."""

    dialect = SQLDialect.DATABRICKS

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "LENGTH": "LENGTH",
            "SUBSTR": "SUBSTRING",
            "NVL": "NVL",
            "IFNULL": "COALESCE",
            "NOW": "CURRENT_TIMESTAMP",
        }

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        # Handle REGEXP
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.REGEXP:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"({left} RLIKE {right})"

        return super().visit_binary_expression(node)


class OracleGenerator(BaseDialectGenerator):
    """SQL generator for Oracle Database."""

    dialect = SQLDialect.ORACLE

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "LENGTH": "LENGTH",
            "SUBSTR": "SUBSTR",
            "IFNULL": "NVL",
            "COALESCE": "COALESCE",
            "NOW": "SYSDATE",
            "CURRENT_TIMESTAMP": "SYSTIMESTAMP",
        }

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        # Handle REGEXP
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.REGEXP:
            left = node.left.accept(self)
            right = node.right.accept(self)
            return f"REGEXP_LIKE({left}, {right})"

        return super().visit_binary_expression(node)


class SQLServerGenerator(BaseDialectGenerator):
    """SQL generator for Microsoft SQL Server."""

    dialect = SQLDialect.SQLSERVER

    def __init__(self, config: DialectConfig | None = None) -> None:
        super().__init__(config)
        self._function_mappings = {
            "LENGTH": "LEN",
            "SUBSTR": "SUBSTRING",
            "NVL": "ISNULL",
            "IFNULL": "ISNULL",
            "NOW": "GETDATE",
        }

    def _quote_identifier(self, name: str) -> str:
        return f"[{name}]"

    def visit_boolean_literal(self, node: BooleanLiteral) -> str:
        return "1" if node.value else "0"

    def visit_binary_expression(self, node: BinaryExpression) -> str:
        # Handle string concatenation
        if isinstance(node.operator, (ArithmeticOp, BinaryOp)):
            if node.operator in (ArithmeticOp.ADD, BinaryOp.ADD):
                # Check if this might be string concatenation
                pass

        # SQL Server doesn't have native REGEXP - would need CLR
        if isinstance(node.operator, ComparisonOp) and node.operator == ComparisonOp.REGEXP:
            raise NotImplementedError(
                "SQL Server does not support native regular expressions. "
                "Consider using LIKE patterns or CLR functions."
            )

        return super().visit_binary_expression(node)


# =============================================================================
# Generator Registry
# =============================================================================


_GENERATOR_REGISTRY: dict[SQLDialect, type[BaseDialectGenerator]] = {
    SQLDialect.POSTGRESQL: PostgreSQLGenerator,
    SQLDialect.MYSQL: MySQLGenerator,
    SQLDialect.SQLITE: SQLiteGenerator,
    SQLDialect.BIGQUERY: BigQueryGenerator,
    SQLDialect.SNOWFLAKE: SnowflakeGenerator,
    SQLDialect.REDSHIFT: RedshiftGenerator,
    SQLDialect.DATABRICKS: DatabricksGenerator,
    SQLDialect.ORACLE: OracleGenerator,
    SQLDialect.SQLSERVER: SQLServerGenerator,
    SQLDialect.GENERIC: BaseDialectGenerator,
}


def get_dialect_generator(
    dialect: SQLDialect,
    config: DialectConfig | None = None,
) -> BaseDialectGenerator:
    """Get a dialect generator instance.

    Args:
        dialect: SQL dialect.
        config: Optional dialect configuration.

    Returns:
        Dialect generator instance.
    """
    generator_class = _GENERATOR_REGISTRY.get(dialect, BaseDialectGenerator)
    return generator_class(config)


def register_dialect_generator(
    dialect: SQLDialect,
    generator_class: type[BaseDialectGenerator],
) -> None:
    """Register a custom dialect generator.

    Args:
        dialect: SQL dialect.
        generator_class: Generator class to register.
    """
    _GENERATOR_REGISTRY[dialect] = generator_class



"""Tests for input adapters."""

import tempfile
from pathlib import Path

import polars as pl
import pytest

from truthound.adapters import to_lazyframe


class TestToLazyframe:
    """Tests for the to_lazyframe function."""

    def test_from_dict(self):
        """Test conversion from dictionary."""
        data = {"a": [1, 2, 3], "b": ["x", "y", "z"]}
        lf = to_lazyframe(data)

        assert isinstance(lf, pl.LazyFrame)
        df = lf.collect()
        assert len(df) == 3
        assert df.columns == ["a", "b"]

    def test_from_polars_dataframe(self):
        """Test conversion from Polars DataFrame."""
        df = pl.DataFrame({"col": [1, 2, 3]})
        lf = to_lazyframe(df)

        assert isinstance(lf, pl.LazyFrame)
        assert lf.collect().equals(df)

    def test_from_polars_lazyframe(self):
        """Test passthrough of Polars LazyFrame."""
        original_lf = pl.DataFrame({"col": [1, 2, 3]}).lazy()
        lf = to_lazyframe(original_lf)

        assert isinstance(lf, pl.LazyFrame)
        assert lf is original_lf

    def test_from_csv_file(self):
        """Test loading from CSV file."""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".csv", delete=False) as f:
            f.write("a,b\n1,x\n2,y\n3,z\n")
            f.flush()

            lf = to_lazyframe(f.name)
            assert isinstance(lf, pl.LazyFrame)

            df = lf.collect()
            assert len(df) == 3
            assert df.columns == ["a", "b"]

            Path(f.name).unlink()

    def test_from_json_file(self):
        """Test loading from JSON file."""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            f.write('[{"a": 1, "b": "x"}, {"a": 2, "b": "y"}]')
            f.flush()

            lf = to_lazyframe(f.name)
            assert isinstance(lf, pl.LazyFrame)

            df = lf.collect()
            assert len(df) == 2

            Path(f.name).unlink()

    def test_from_parquet_file(self):
        """Test loading from Parquet file."""
        with tempfile.NamedTemporaryFile(suffix=".parquet", delete=False) as f:
            df = pl.DataFrame({"a": [1, 2, 3], "b": ["x", "y", "z"]})
            df.write_parquet(f.name)

            lf = to_lazyframe(f.name)
            assert isinstance(lf, pl.LazyFrame)

            loaded_df = lf.collect()
            assert loaded_df.equals(df)

            Path(f.name).unlink()

    def test_file_not_found(self):
        """Test error handling for non-existent file."""
        with pytest.raises(FileNotFoundError):
            to_lazyframe("/nonexistent/file.csv")

    def test_unsupported_extension(self):
        """Test error handling for unsupported file extension."""
        with tempfile.NamedTemporaryFile(suffix=".xyz", delete=False) as f:
            f.write(b"test")
            f.flush()

            with pytest.raises(ValueError, match="Unsupported file extension"):
                to_lazyframe(f.name)

            Path(f.name).unlink()

    def test_unsupported_type(self):
        """Test error handling for unsupported input type."""
        with pytest.raises(ValueError, match="Unsupported input type"):
            to_lazyframe(12345)


class TestPandasIntegration:
    """Tests for pandas DataFrame integration."""

    def test_from_pandas_dataframe(self):
        """Test conversion from pandas DataFrame."""
        pytest.importorskip("pandas")
        import pandas as pd

        pdf = pd.DataFrame({"a": [1, 2, 3], "b": ["x", "y", "z"]})
        lf = to_lazyframe(pdf)

        assert isinstance(lf, pl.LazyFrame)
        df = lf.collect()
        assert len(df) == 3
        assert df.columns == ["a", "b"]



"""Metrics collection system for Truthound.

This module provides a flexible metrics collection framework with support
for multiple backend exporters (Prometheus, StatsD, OpenTelemetry).

Metric Types:
    - Counter: Monotonically increasing value (e.g., request count)
    - Gauge: Point-in-time value (e.g., active connections)
    - Histogram: Distribution of values with buckets
    - Summary: Distribution with quantiles

Design Principles:
    1. Label-based: Dimensional metrics with key-value labels
    2. Backend agnostic: Same API for all exporters
    3. Lazy registration: Metrics created on first use
    4. Thread-safe: All operations are thread-safe
"""

from __future__ import annotations

import threading
import time
from abc import ABC, abstractmethod
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from typing import Any, Callable, Iterator, TypeVar


# =============================================================================
# Metric Types
# =============================================================================


class MetricType(Enum):
    """Types of metrics."""

    COUNTER = "counter"
    GAUGE = "gauge"
    HISTOGRAM = "histogram"
    SUMMARY = "summary"


@dataclass(frozen=True)
class MetricKey:
    """Unique identifier for a metric with labels."""

    name: str
    labels: tuple[tuple[str, str], ...]

    @classmethod
    def create(cls, name: str, labels: dict[str, str] | None = None) -> "MetricKey":
        """Create metric key from name and labels."""
        label_tuple = tuple(sorted((labels or {}).items()))
        return cls(name=name, labels=label_tuple)


@dataclass
class MetricMetadata:
    """Metadata about a metric."""

    name: str
    type: MetricType
    description: str
    unit: str = ""
    label_names: tuple[str, ...] = ()


# =============================================================================
# Metric Base Class
# =============================================================================


class Metric(ABC):
    """Abstract base class for metrics.

    All metrics support labels (dimensions) for multi-dimensional analysis.
    """

    def __init__(
        self,
        name: str,
        description: str = "",
        *,
        unit: str = "",
        label_names: tuple[str, ...] | list[str] = (),
    ) -> None:
        """Initialize metric.

        Args:
            name: Metric name (should be lowercase with underscores).
            description: Human-readable description.
            unit: Unit of measurement.
            label_names: Names of labels for this metric.
        """
        self._name = name
        self._description = description
        self._unit = unit
        self._label_names = tuple(label_names)
        self._lock = threading.Lock()

    @property
    def name(self) -> str:
        """Get metric name."""
        return self._name

    @property
    def description(self) -> str:
        """Get metric description."""
        return self._description

    @property
    @abstractmethod
    def type(self) -> MetricType:
        """Get metric type."""
        pass

    @property
    def metadata(self) -> MetricMetadata:
        """Get metric metadata."""
        return MetricMetadata(
            name=self._name,
            type=self.type,
            description=self._description,
            unit=self._unit,
            label_names=self._label_names,
        )

    def _validate_labels(self, labels: dict[str, str]) -> None:
        """Validate label names match expected."""
        if set(labels.keys()) != set(self._label_names):
            expected = set(self._label_names)
            actual = set(labels.keys())
            raise ValueError(
                f"Label mismatch for '{self._name}': "
                f"expected {expected}, got {actual}"
            )

    @abstractmethod
    def collect(self) -> list[tuple[dict[str, str], float]]:
        """Collect all metric values with labels.

        Returns:
            List of (labels, value) tuples.
        """
        pass


# =============================================================================
# Counter
# =============================================================================


class Counter(Metric):
    """Monotonically increasing counter.

    Counters only go up (and reset to zero on restart).
    Use for: request counts, errors, completed tasks.

    Example:
        >>> requests = Counter("http_requests_total", "Total HTTP requests")
        >>> requests.inc()
        >>> requests.add(5)
        >>> requests.labels(method="GET", status="200").inc()
    """

    def __init__(
        self,
        name: str,
        description: str = "",
        *,
        labels: tuple[str, ...] | list[str] = (),
        **kwargs: Any,
    ) -> None:
        """Initialize counter.

        Args:
            name: Counter name.
            description: Human-readable description.
            labels: Label names for this counter.
            **kwargs: Additional arguments for Metric.
        """
        super().__init__(name, description, label_names=labels, **kwargs)
        self._values: dict[MetricKey, float] = {}

    @property
    def type(self) -> MetricType:
        return MetricType.COUNTER

    def inc(self, value: float = 1.0, **labels: str) -> None:
        """Increment counter.

        Args:
            value: Amount to increment (must be positive).
            **labels: Label values.
        """
        if value < 0:
            raise ValueError("Counter can only be incremented")
        self.add(value, **labels)

    def add(self, value: float, **labels: str) -> None:
        """Add to counter.

        Args:
            value: Amount to add (must be positive).
            **labels: Label values.
        """
        if value < 0:
            raise ValueError("Counter can only increase")

        if self._label_names:
            self._validate_labels(labels)

        key = MetricKey.create(self._name, labels)

        with self._lock:
            self._values[key] = self._values.get(key, 0.0) + value

    def labels(self, **labels: str) -> "LabeledCounter":
        """Get counter with specific labels.

        Args:
            **labels: Label values.

        Returns:
            LabeledCounter for the specific label set.
        """
        return LabeledCounter(self, labels)

    def get(self, **labels: str) -> float:
        """Get current counter value.

        Args:
            **labels: Label values.

        Returns:
            Current value.
        """
        key = MetricKey.create(self._name, labels)
        with self._lock:
            return self._values.get(key, 0.0)

    def collect(self) -> list[tuple[dict[str, str], float]]:
        """Collect all counter values."""
        with self._lock:
            return [
                (dict(key.labels), value)
                for key, value in self._values.items()
            ]


class LabeledCounter:
    """Counter with pre-set labels."""

    def __init__(self, counter: Counter, labels: dict[str, str]) -> None:
        self._counter = counter
        self._labels = labels

    def inc(self, value: float = 1.0) -> None:
        """Increment counter."""
        self._counter.inc(value, **self._labels)

    def add(self, value: float) -> None:
        """Add to counter."""
        self._counter.add(value, **self._labels)


# =============================================================================
# Gauge
# =============================================================================


class Gauge(Metric):
    """Point-in-time value that can go up or down.

    Use for: temperature, queue size, memory usage.

    Example:
        >>> temperature = Gauge("temperature_celsius", "Current temperature")
        >>> temperature.set(23.5)
        >>> temperature.inc()
        >>> temperature.dec(0.5)
    """

    def __init__(
        self,
        name: str,
        description: str = "",
        *,
        labels: tuple[str, ...] | list[str] = (),
        **kwargs: Any,
    ) -> None:
        """Initialize gauge."""
        super().__init__(name, description, label_names=labels, **kwargs)
        self._values: dict[MetricKey, float] = {}

    @property
    def type(self) -> MetricType:
        return MetricType.GAUGE

    def set(self, value: float, **labels: str) -> None:
        """Set gauge value.

        Args:
            value: New value.
            **labels: Label values.
        """
        if self._label_names:
            self._validate_labels(labels)

        key = MetricKey.create(self._name, labels)

        with self._lock:
            self._values[key] = value

    def inc(self, value: float = 1.0, **labels: str) -> None:
        """Increment gauge.

        Args:
            value: Amount to increment.
            **labels: Label values.
        """
        key = MetricKey.create(self._name, labels if self._label_names else {})

        with self._lock:
            self._values[key] = self._values.get(key, 0.0) + value

    def dec(self, value: float = 1.0, **labels: str) -> None:
        """Decrement gauge.

        Args:
            value: Amount to decrement.
            **labels: Label values.
        """
        self.inc(-value, **labels)

    def get(self, **labels: str) -> float:
        """Get current gauge value."""
        key = MetricKey.create(self._name, labels)
        with self._lock:
            return self._values.get(key, 0.0)

    def labels(self, **labels: str) -> "LabeledGauge":
        """Get gauge with specific labels."""
        return LabeledGauge(self, labels)

    @contextmanager
    def track_inprogress(self, **labels: str) -> Iterator[None]:
        """Track in-progress operations.

        Increments on entry, decrements on exit.
        """
        self.inc(**labels)
        try:
            yield
        finally:
            self.dec(**labels)

    def set_to_current_time(self, **labels: str) -> None:
        """Set gauge to current Unix timestamp."""
        self.set(time.time(), **labels)

    def collect(self) -> list[tuple[dict[str, str], float]]:
        """Collect all gauge values."""
        with self._lock:
            return [
                (dict(key.labels), value)
                for key, value in self._values.items()
            ]


class LabeledGauge:
    """Gauge with pre-set labels."""

    def __init__(self, gauge: Gauge, labels: dict[str, str]) -> None:
        self._gauge = gauge
        self._labels = labels

    def set(self, value: float) -> None:
        """Set gauge value."""
        self._gauge.set(value, **self._labels)

    def inc(self, value: float = 1.0) -> None:
        """Increment gauge."""
        self._gauge.inc(value, **self._labels)

    def dec(self, value: float = 1.0) -> None:
        """Decrement gauge."""
        self._gauge.dec(value, **self._labels)


# =============================================================================
# Histogram
# =============================================================================


class Histogram(Metric):
    """Distribution of values with configurable buckets.

    Use for: request latency, response sizes.

    Example:
        >>> latency = Histogram(
        ...     "request_duration_seconds",
        ...     "Request latency",
        ...     buckets=[0.01, 0.05, 0.1, 0.5, 1.0, 5.0],
        ... )
        >>> latency.observe(0.42)
        >>> with latency.time():
        ...     process_request()
    """

    DEFAULT_BUCKETS = (
        0.005, 0.01, 0.025, 0.05, 0.075, 0.1, 0.25, 0.5,
        0.75, 1.0, 2.5, 5.0, 7.5, 10.0, float("inf"),
    )

    def __init__(
        self,
        name: str,
        description: str = "",
        *,
        buckets: tuple[float, ...] | list[float] | None = None,
        labels: tuple[str, ...] | list[str] = (),
        **kwargs: Any,
    ) -> None:
        """Initialize histogram.

        Args:
            name: Histogram name.
            description: Human-readable description.
            buckets: Upper bounds for buckets.
            labels: Label names.
            **kwargs: Additional arguments.
        """
        super().__init__(name, description, label_names=labels, **kwargs)
        self._buckets = tuple(sorted(buckets or self.DEFAULT_BUCKETS))

        # Ensure +Inf is included
        if self._buckets[-1] != float("inf"):
            self._buckets = (*self._buckets, float("inf"))

        # Storage: key -> (bucket_counts, sum, count)
        self._data: dict[MetricKey, tuple[list[int], float, int]] = {}

    @property
    def type(self) -> MetricType:
        return MetricType.HISTOGRAM

    @property
    def buckets(self) -> tuple[float, ...]:
        """Get bucket boundaries."""
        return self._buckets

    def _get_data(self, key: MetricKey) -> tuple[list[int], float, int]:
        """Get or create data for a key."""
        if key not in self._data:
            self._data[key] = ([0] * len(self._buckets), 0.0, 0)
        return self._data[key]

    def observe(self, value: float, **labels: str) -> None:
        """Observe a value.

        Args:
            value: Observed value.
            **labels: Label values.
        """
        if self._label_names:
            self._validate_labels(labels)

        key = MetricKey.create(self._name, labels)

        with self._lock:
            bucket_counts, total_sum, count = self._get_data(key)

            # Find the bucket this value belongs to (first bucket where value <= bound)
            # Only increment that one bucket - cumulative is computed in collect()
            for i, bound in enumerate(self._buckets):
                if value <= bound:
                    bucket_counts[i] += 1
                    break

            # Update sum and count
            self._data[key] = (bucket_counts, total_sum + value, count + 1)

    @contextmanager
    def time(self, **labels: str) -> Iterator[None]:
        """Context manager to measure duration.

        Args:
            **labels: Label values.
        """
        start = time.perf_counter()
        try:
            yield
        finally:
            duration = time.perf_counter() - start
            self.observe(duration, **labels)

    def labels(self, **labels: str) -> "LabeledHistogram":
        """Get histogram with specific labels."""
        return LabeledHistogram(self, labels)

    def collect(self) -> list[tuple[dict[str, str], dict[str, Any]]]:
        """Collect histogram data.

        Returns list of (labels, data) where data contains:
        - buckets: dict mapping bound to cumulative count
        - sum: total of all observed values
        - count: number of observations
        """
        with self._lock:
            results = []
            for key, (bucket_counts, total_sum, count) in self._data.items():
                # Calculate cumulative counts
                cumulative = []
                running = 0
                for c in bucket_counts:
                    running += c
                    cumulative.append(running)

                data = {
                    "buckets": {
                        str(bound): cum
                        for bound, cum in zip(self._buckets, cumulative)
                    },
                    "sum": total_sum,
                    "count": count,
                }
                results.append((dict(key.labels), data))
            return results


class LabeledHistogram:
    """Histogram with pre-set labels."""

    def __init__(self, histogram: Histogram, labels: dict[str, str]) -> None:
        self._histogram = histogram
        self._labels = labels

    def observe(self, value: float) -> None:
        """Observe a value."""
        self._histogram.observe(value, **self._labels)

    @contextmanager
    def time(self) -> Iterator[None]:
        """Time a block of code."""
        with self._histogram.time(**self._labels):
            yield


# =============================================================================
# Summary
# =============================================================================


class Summary(Metric):
    """Summary with streaming quantiles.

    Similar to Histogram but calculates quantiles on the client side.
    Use when you need specific percentiles (p50, p90, p99).

    Note: This is a simplified implementation that keeps recent samples.
    For production use, consider a proper streaming quantile algorithm.
    """

    def __init__(
        self,
        name: str,
        description: str = "",
        *,
        quantiles: tuple[float, ...] = (0.5, 0.9, 0.99),
        max_samples: int = 1000,
        labels: tuple[str, ...] | list[str] = (),
        **kwargs: Any,
    ) -> None:
        """Initialize summary.

        Args:
            name: Summary name.
            description: Human-readable description.
            quantiles: Quantiles to calculate (0.0-1.0).
            max_samples: Maximum samples to keep.
            labels: Label names.
            **kwargs: Additional arguments.
        """
        super().__init__(name, description, label_names=labels, **kwargs)
        self._quantiles = quantiles
        self._max_samples = max_samples
        self._data: dict[MetricKey, tuple[list[float], float, int]] = {}

    @property
    def type(self) -> MetricType:
        return MetricType.SUMMARY

    def observe(self, value: float, **labels: str) -> None:
        """Observe a value.

        Args:
            value: Observed value.
            **labels: Label values.
        """
        if self._label_names:
            self._validate_labels(labels)

        key = MetricKey.create(self._name, labels)

        with self._lock:
            if key not in self._data:
                self._data[key] = ([], 0.0, 0)

            samples, total_sum, count = self._data[key]

            # Add sample (circular buffer behavior)
            if len(samples) >= self._max_samples:
                samples.pop(0)
            samples.append(value)

            self._data[key] = (samples, total_sum + value, count + 1)

    @contextmanager
    def time(self, **labels: str) -> Iterator[None]:
        """Time a block of code."""
        start = time.perf_counter()
        try:
            yield
        finally:
            self.observe(time.perf_counter() - start, **labels)

    def labels(self, **labels: str) -> "LabeledSummary":
        """Get summary with specific labels."""
        return LabeledSummary(self, labels)

    def collect(self) -> list[tuple[dict[str, str], dict[str, Any]]]:
        """Collect summary data with quantiles."""
        with self._lock:
            results = []
            for key, (samples, total_sum, count) in self._data.items():
                quantile_values = {}
                if samples:
                    sorted_samples = sorted(samples)
                    n = len(sorted_samples)
                    for q in self._quantiles:
                        idx = int(q * (n - 1))
                        quantile_values[f"p{int(q*100)}"] = sorted_samples[idx]

                data = {
                    "quantiles": quantile_values,
                    "sum": total_sum,
                    "count": count,
                }
                results.append((dict(key.labels), data))
            return results


class LabeledSummary:
    """Summary with pre-set labels."""

    def __init__(self, summary: Summary, labels: dict[str, str]) -> None:
        self._summary = summary
        self._labels = labels

    def observe(self, value: float) -> None:
        """Observe a value."""
        self._summary.observe(value, **self._labels)


# =============================================================================
# Metrics Registry
# =============================================================================


class MetricsRegistry:
    """Central registry for all metrics.

    Ensures unique metric names and provides collection.
    """

    def __init__(self) -> None:
        """Initialize registry."""
        self._metrics: dict[str, Metric] = {}
        self._lock = threading.Lock()

    def register(self, metric: Metric) -> Metric:
        """Register a metric.

        Args:
            metric: Metric to register.

        Returns:
            The registered metric.

        Raises:
            ValueError: If metric name already registered with different type.
        """
        with self._lock:
            if metric.name in self._metrics:
                existing = self._metrics[metric.name]
                if existing.type != metric.type:
                    raise ValueError(
                        f"Metric '{metric.name}' already registered "
                        f"as {existing.type.value}"
                    )
                return existing
            self._metrics[metric.name] = metric
            return metric

    def get(self, name: str) -> Metric | None:
        """Get a registered metric by name."""
        return self._metrics.get(name)

    def unregister(self, name: str) -> bool:
        """Unregister a metric.

        Args:
            name: Metric name.

        Returns:
            True if unregistered, False if not found.
        """
        with self._lock:
            if name in self._metrics:
                del self._metrics[name]
                return True
            return False

    def collect_all(self) -> dict[str, Any]:
        """Collect all metrics.

        Returns:
            Dictionary mapping metric names to their collected data.
        """
        with self._lock:
            result = {}
            for name, metric in self._metrics.items():
                result[name] = {
                    "type": metric.type.value,
                    "description": metric.description,
                    "data": metric.collect(),
                }
            return result

    def clear(self) -> None:
        """Clear all registered metrics."""
        with self._lock:
            self._metrics.clear()


# =============================================================================
# Metrics Collector
# =============================================================================


class MetricsCollector:
    """High-level interface for creating and managing metrics.

    MetricsCollector provides a fluent API for creating metrics and
    manages their registration in the registry.

    Example:
        >>> collector = MetricsCollector()
        >>> requests = collector.counter(
        ...     "http_requests_total",
        ...     "Total HTTP requests",
        ...     labels=["method", "status"],
        ... )
        >>> latency = collector.histogram(
        ...     "request_duration_seconds",
        ...     "Request latency",
        ... )
    """

    def __init__(self, registry: MetricsRegistry | None = None) -> None:
        """Initialize collector.

        Args:
            registry: Metrics registry (creates new if None).
        """
        self._registry = registry or MetricsRegistry()

    @property
    def registry(self) -> MetricsRegistry:
        """Get the metrics registry."""
        return self._registry

    def counter(
        self,
        name: str,
        description: str = "",
        *,
        labels: list[str] | tuple[str, ...] = (),
        **kwargs: Any,
    ) -> Counter:
        """Create or get a counter.

        Args:
            name: Counter name.
            description: Human-readable description.
            labels: Label names.
            **kwargs: Additional arguments.

        Returns:
            Counter instance.
        """
        counter = Counter(name, description, labels=labels, **kwargs)
        return self._registry.register(counter)  # type: ignore

    def gauge(
        self,
        name: str,
        description: str = "",
        *,
        labels: list[str] | tuple[str, ...] = (),
        **kwargs: Any,
    ) -> Gauge:
        """Create or get a gauge.

        Args:
            name: Gauge name.
            description: Human-readable description.
            labels: Label names.
            **kwargs: Additional arguments.

        Returns:
            Gauge instance.
        """
        gauge = Gauge(name, description, labels=labels, **kwargs)
        return self._registry.register(gauge)  # type: ignore

    def histogram(
        self,
        name: str,
        description: str = "",
        *,
        buckets: list[float] | tuple[float, ...] | None = None,
        labels: list[str] | tuple[str, ...] = (),
        **kwargs: Any,
    ) -> Histogram:
        """Create or get a histogram.

        Args:
            name: Histogram name.
            description: Human-readable description.
            buckets: Bucket boundaries.
            labels: Label names.
            **kwargs: Additional arguments.

        Returns:
            Histogram instance.
        """
        histogram = Histogram(
            name, description, buckets=buckets, labels=labels, **kwargs
        )
        return self._registry.register(histogram)  # type: ignore

    def summary(
        self,
        name: str,
        description: str = "",
        *,
        quantiles: tuple[float, ...] = (0.5, 0.9, 0.99),
        labels: list[str] | tuple[str, ...] = (),
        **kwargs: Any,
    ) -> Summary:
        """Create or get a summary.

        Args:
            name: Summary name.
            description: Human-readable description.
            quantiles: Quantiles to calculate.
            labels: Label names.
            **kwargs: Additional arguments.

        Returns:
            Summary instance.
        """
        summary = Summary(
            name, description, quantiles=quantiles, labels=labels, **kwargs
        )
        return self._registry.register(summary)  # type: ignore


# =============================================================================
# Metrics Exporters
# =============================================================================


class MetricsExporter(ABC):
    """Abstract base class for metrics exporters."""

    @abstractmethod
    def export(self, registry: MetricsRegistry) -> str:
        """Export metrics from registry.

        Args:
            registry: Metrics registry to export.

        Returns:
            Exported metrics as string.
        """
        pass

    @abstractmethod
    def push(self, registry: MetricsRegistry, endpoint: str) -> bool:
        """Push metrics to remote endpoint.

        Args:
            registry: Metrics registry to push.
            endpoint: Remote endpoint URL.

        Returns:
            True if successful.
        """
        pass


class PrometheusExporter(MetricsExporter):
    """Export metrics in Prometheus text format.

    Generates output compatible with Prometheus text exposition format.

    Example output:
        # HELP http_requests_total Total HTTP requests
        # TYPE http_requests_total counter
        http_requests_total{method="GET",status="200"} 1234
    """

    def export(self, registry: MetricsRegistry) -> str:
        """Export metrics in Prometheus format."""
        lines = []
        data = registry.collect_all()

        for name, info in data.items():
            metric_type = info["type"]
            description = info["description"]
            metric_data = info["data"]

            # HELP and TYPE lines
            lines.append(f"# HELP {name} {description}")
            lines.append(f"# TYPE {name} {metric_type}")

            # Metric lines
            for labels, value in metric_data:
                label_str = self._format_labels(labels)

                if metric_type == "histogram":
                    # Histogram has special format
                    for bound, count in value["buckets"].items():
                        bound_label = f'{label_str},le="{bound}"' if label_str else f'le="{bound}"'
                        lines.append(f"{name}_bucket{{{bound_label}}} {count}")
                    sum_labels = f"{{{label_str}}}" if label_str else ""
                    lines.append(f"{name}_sum{sum_labels} {value['sum']}")
                    lines.append(f"{name}_count{sum_labels} {value['count']}")

                elif metric_type == "summary":
                    for quantile_name, quantile_value in value.get("quantiles", {}).items():
                        q = quantile_name[1:]  # Remove 'p' prefix
                        q_label = f'{label_str},quantile="{int(q)/100}"' if label_str else f'quantile="{int(q)/100}"'
                        lines.append(f"{name}{{{q_label}}} {quantile_value}")
                    sum_labels = f"{{{label_str}}}" if label_str else ""
                    lines.append(f"{name}_sum{sum_labels} {value['sum']}")
                    lines.append(f"{name}_count{sum_labels} {value['count']}")

                else:
                    # Counter and Gauge
                    if label_str:
                        lines.append(f"{name}{{{label_str}}} {value}")
                    else:
                        lines.append(f"{name} {value}")

            lines.append("")  # Empty line between metrics

        return "\n".join(lines)

    def _format_labels(self, labels: dict[str, str]) -> str:
        """Format labels for Prometheus."""
        if not labels:
            return ""
        parts = [f'{k}="{v}"' for k, v in sorted(labels.items())]
        return ",".join(parts)

    def push(self, registry: MetricsRegistry, endpoint: str) -> bool:
        """Push metrics to Prometheus Pushgateway."""
        import urllib.request
        import urllib.error

        data = self.export(registry).encode("utf-8")

        try:
            request = urllib.request.Request(
                endpoint,
                data=data,
                method="POST",
                headers={"Content-Type": "text/plain"},
            )
            with urllib.request.urlopen(request, timeout=30):
                return True
        except urllib.error.URLError:
            return False


class StatsDExporter(MetricsExporter):
    """Export metrics in StatsD format.

    Sends metrics via UDP to a StatsD server.
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 8125,
        prefix: str = "",
    ) -> None:
        """Initialize StatsD exporter.

        Args:
            host: StatsD server host.
            port: StatsD server port.
            prefix: Metric name prefix.
        """
        self._host = host
        self._port = port
        self._prefix = prefix

    def export(self, registry: MetricsRegistry) -> str:
        """Export metrics in StatsD format."""
        lines = []
        data = registry.collect_all()

        for name, info in data.items():
            metric_type = info["type"]
            metric_data = info["data"]

            full_name = f"{self._prefix}{name}" if self._prefix else name

            for labels, value in metric_data:
                label_suffix = self._format_labels(labels)
                metric_name = f"{full_name}{label_suffix}"

                if metric_type == "counter":
                    lines.append(f"{metric_name}:{value}|c")
                elif metric_type == "gauge":
                    lines.append(f"{metric_name}:{value}|g")
                elif metric_type == "histogram":
                    # Send as timing
                    lines.append(f"{metric_name}:{value['sum']/max(value['count'], 1)*1000}|ms")
                elif metric_type == "summary":
                    lines.append(f"{metric_name}:{value['sum']/max(value['count'], 1)*1000}|ms")

        return "\n".join(lines)

    def _format_labels(self, labels: dict[str, str]) -> str:
        """Format labels as suffix."""
        if not labels:
            return ""
        parts = [f".{k}_{v}" for k, v in sorted(labels.items())]
        return "".join(parts)

    def push(self, registry: MetricsRegistry, endpoint: str | None = None) -> bool:
        """Send metrics to StatsD server via UDP."""
        import socket

        data = self.export(registry)

        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.sendto(data.encode("utf-8"), (self._host, self._port))
            sock.close()
            return True
        except socket.error:
            return False


class InMemoryExporter(MetricsExporter):
    """In-memory exporter for testing.

    Stores exported metrics for inspection.
    """

    def __init__(self) -> None:
        """Initialize in-memory exporter."""
        self._exports: list[dict[str, Any]] = []

    @property
    def exports(self) -> list[dict[str, Any]]:
        """Get all exports."""
        return self._exports

    def export(self, registry: MetricsRegistry) -> str:
        """Export metrics to memory."""
        import json
        data = registry.collect_all()
        self._exports.append(data)
        return json.dumps(data, default=str)

    def push(self, registry: MetricsRegistry, endpoint: str) -> bool:
        """Push to memory (always succeeds)."""
        self.export(registry)
        return True

    def clear(self) -> None:
        """Clear stored exports."""
        self._exports.clear()


# =============================================================================
# Global Metrics
# =============================================================================

_global_collector: MetricsCollector | None = None
_lock = threading.Lock()


def get_metrics() -> MetricsCollector:
    """Get the global metrics collector.

    Returns:
        Global MetricsCollector instance.
    """
    global _global_collector

    with _lock:
        if _global_collector is None:
            _global_collector = MetricsCollector()
        return _global_collector


def set_metrics(collector: MetricsCollector) -> None:
    """Set the global metrics collector.

    Args:
        collector: MetricsCollector to use globally.
    """
    global _global_collector

    with _lock:
        _global_collector = collector


def configure_metrics(
    *,
    exporter: MetricsExporter | None = None,
    registry: MetricsRegistry | None = None,
) -> MetricsCollector:
    """Configure global metrics.

    Args:
        exporter: Exporter to use.
        registry: Registry to use.

    Returns:
        Configured MetricsCollector.
    """
    collector = MetricsCollector(registry=registry)
    set_metrics(collector)
    return collector



"""Metric collectors for monitoring.

Collectors gather metrics from various sources (in-memory, Redis, Prometheus)
and expose them through a unified protocol.
"""

from truthound.checkpoint.monitoring.collectors.base import BaseCollector
from truthound.checkpoint.monitoring.collectors.memory_collector import InMemoryCollector
from truthound.checkpoint.monitoring.collectors.redis_collector import RedisCollector
from truthound.checkpoint.monitoring.collectors.prometheus_collector import PrometheusCollector

__all__ = [
    "BaseCollector",
    "InMemoryCollector",
    "RedisCollector",
    "PrometheusCollector",
]



"""Tests for drift detection."""

import polars as pl
import pytest

import truthound as th
from truthound.drift.detectors import (
    ChiSquareDetector,
    DriftLevel,
    JensenShannonDetector,
    KSTestDetector,
    PSIDetector,
)


class TestKSTestDetector:
    """Tests for Kolmogorov-Smirnov test detector."""

    def test_no_drift_same_distribution(self):
        """Test that identical distributions show no drift."""
        data = pl.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        detector = KSTestDetector()
        result = detector.detect(data, data)

        assert result.statistic == 0.0
        assert not result.drifted

    def test_detects_drift_different_distributions(self):
        """Test detection of drift between different distributions."""
        baseline = pl.Series([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
        current = pl.Series([50, 60, 70, 80, 90, 100, 110, 120, 130, 140])

        detector = KSTestDetector()
        result = detector.detect(baseline, current)

        assert result.statistic > 0.5
        assert result.drifted
        assert result.level in (DriftLevel.MEDIUM, DriftLevel.HIGH)

    def test_handles_empty_data(self):
        """Test handling of empty series."""
        baseline = pl.Series([1, 2, 3])
        current = pl.Series([], dtype=pl.Int64)

        detector = KSTestDetector()
        result = detector.detect(baseline, current)

        assert not result.drifted
        assert "Insufficient data" in result.details


class TestPSIDetector:
    """Tests for Population Stability Index detector."""

    def test_no_drift_same_distribution(self):
        """Test that similar distributions show low PSI."""
        data = pl.Series(list(range(100)))
        detector = PSIDetector()
        result = detector.detect(data, data)

        assert result.statistic < 0.1
        assert not result.drifted

    def test_detects_drift_shifted_distribution(self):
        """Test detection of shifted distribution."""
        baseline = pl.Series(list(range(100)))
        current = pl.Series(list(range(50, 150)))

        detector = PSIDetector()
        result = detector.detect(baseline, current)

        # Shift should cause some PSI but might not exceed threshold
        assert result.statistic > 0


class TestChiSquareDetector:
    """Tests for Chi-square detector."""

    def test_no_drift_same_categories(self):
        """Test that identical categorical distributions show no drift."""
        data = pl.Series(["a", "b", "c", "a", "b", "c"] * 10)
        detector = ChiSquareDetector()
        result = detector.detect(data, data)

        assert result.statistic < 0.1
        assert not result.drifted

    def test_detects_drift_different_categories(self):
        """Test detection of categorical drift."""
        baseline = pl.Series(["a", "a", "a", "b", "b", "c"] * 10)
        current = pl.Series(["a", "b", "b", "b", "c", "c"] * 10)

        detector = ChiSquareDetector()
        result = detector.detect(baseline, current)

        # Distribution changed, should detect drift
        assert result.statistic > 0

    def test_detects_new_category(self):
        """Test detection of new category appearing."""
        baseline = pl.Series(["a", "b", "c"] * 20)
        current = pl.Series(["a", "b", "c", "d"] * 15)

        detector = ChiSquareDetector()
        result = detector.detect(baseline, current)

        assert result.drifted


class TestJensenShannonDetector:
    """Tests for Jensen-Shannon divergence detector."""

    def test_no_drift_same_distribution(self):
        """Test that identical distributions have zero JS divergence."""
        data = pl.Series(list(range(100)))
        detector = JensenShannonDetector()
        result = detector.detect(data, data)

        assert result.statistic < 0.01
        assert not result.drifted

    def test_detects_drift(self):
        """Test detection of distribution drift."""
        baseline = pl.Series(list(range(0, 100)))
        current = pl.Series(list(range(80, 180)))

        detector = JensenShannonDetector()
        result = detector.detect(baseline, current)

        assert result.statistic > 0


class TestCompare:
    """Tests for th.compare() function."""

    def test_compare_identical_data(self):
        """Test comparing identical datasets."""
        data = {"value": [1, 2, 3, 4, 5], "category": ["a", "b", "c", "a", "b"]}

        report = th.compare(data, data)

        assert not report.has_drift
        assert len(report.get_drifted_columns()) == 0

    def test_compare_drifted_numeric(self):
        """Test detecting drift in numeric column."""
        baseline = {"value": list(range(100))}
        current = {"value": list(range(500, 600))}

        report = th.compare(baseline, current)

        assert report.has_drift
        assert "value" in report.get_drifted_columns()

    def test_compare_drifted_categorical(self):
        """Test detecting drift in categorical column."""
        baseline = {"status": ["active"] * 50 + ["inactive"] * 50}
        current = {"status": ["active"] * 10 + ["inactive"] * 90}

        report = th.compare(baseline, current)

        # Should detect the distribution shift
        assert len(report.columns) == 1

    def test_compare_specific_columns(self):
        """Test comparing specific columns only."""
        baseline = {"a": [1, 2, 3], "b": [4, 5, 6], "c": [7, 8, 9]}
        current = {"a": [100, 200, 300], "b": [4, 5, 6], "c": [700, 800, 900]}

        report = th.compare(baseline, current, columns=["a", "b"])

        assert len(report.columns) == 2
        column_names = [c.column for c in report.columns]
        assert "a" in column_names
        assert "b" in column_names
        assert "c" not in column_names

    def test_compare_specific_method(self):
        """Test using specific detection method."""
        baseline = {"value": list(range(100))}
        current = {"value": list(range(50, 150))}

        report_ks = th.compare(baseline, current, method="ks")
        report_psi = th.compare(baseline, current, method="psi")

        # Both should have analyzed the column
        assert len(report_ks.columns) == 1
        assert len(report_psi.columns) == 1

        # Methods should be different
        assert report_ks.columns[0].result.method == "ks_test"
        assert report_psi.columns[0].result.method == "psi"

    def test_compare_with_threshold(self):
        """Test using custom threshold."""
        baseline = {"value": list(range(100))}
        current = {"value": list(range(10, 110))}

        # Loose threshold
        report_loose = th.compare(baseline, current, method="psi", threshold=0.5)
        # Strict threshold
        report_strict = th.compare(baseline, current, method="psi", threshold=0.01)

        # Strict threshold more likely to detect drift
        assert report_strict.columns[0].result.threshold < report_loose.columns[0].result.threshold

    def test_report_json_output(self):
        """Test JSON output of drift report."""
        baseline = {"value": [1, 2, 3, 4, 5]}
        current = {"value": [10, 20, 30, 40, 50]}

        report = th.compare(baseline, current)
        json_str = report.to_json()

        import json
        data = json.loads(json_str)

        assert "baseline_source" in data
        assert "current_source" in data
        assert "columns" in data
        assert len(data["columns"]) == 1


class TestDriftReport:
    """Tests for DriftReport."""

    def test_has_drift_property(self):
        """Test has_drift property."""
        baseline = {"value": list(range(100))}
        current = {"value": list(range(500, 600))}

        report = th.compare(baseline, current)

        assert report.has_drift is True

    def test_get_drifted_columns(self):
        """Test getting list of drifted columns."""
        baseline = {
            "stable": list(range(100)),
            "drifted": list(range(100)),
        }
        current = {
            "stable": list(range(100)),
            "drifted": list(range(500, 600)),
        }

        report = th.compare(baseline, current)
        drifted = report.get_drifted_columns()

        assert "drifted" in drifted


class TestSampling:
    """Tests for sampling optimization."""

    def test_compare_with_sampling(self):
        """Test comparison with sampling for large datasets."""
        baseline = {"value": list(range(10000))}
        current = {"value": list(range(5000, 15000))}

        # Compare with sampling
        report = th.compare(baseline, current, sample_size=1000)

        assert len(report.columns) == 1
        # Should still detect drift even with sampling
        assert report.has_drift

    def test_sampling_deterministic(self):
        """Test that sampling is deterministic with seed."""
        baseline = {"value": list(range(10000))}
        current = {"value": list(range(5000, 15000))}

        report1 = th.compare(baseline, current, sample_size=1000)
        report2 = th.compare(baseline, current, sample_size=1000)

        # Same sampling should give same results
        assert report1.columns[0].result.statistic == report2.columns[0].result.statistic



"""Tests for plugin base classes and types."""

from __future__ import annotations

import pytest
from dataclasses import dataclass

from truthound.plugins import (
    Plugin,
    PluginConfig,
    PluginInfo,
    PluginType,
    PluginState,
    PluginError,
    PluginLoadError,
    PluginNotFoundError,
    PluginDependencyError,
    PluginCompatibilityError,
    ValidatorPlugin,
    ReporterPlugin,
    HookPlugin,
)


# =============================================================================
# Test Fixtures
# =============================================================================


class SimpleTestPlugin(Plugin[PluginConfig]):
    """Simple test plugin implementation."""

    def __init__(self, config: PluginConfig | None = None):
        super().__init__(config)
        self.setup_called = False
        self.teardown_called = False
        self.register_called = False

    @property
    def info(self) -> PluginInfo:
        return PluginInfo(
            name="simple-test",
            version="1.0.0",
            plugin_type=PluginType.CUSTOM,
            description="A simple test plugin",
            author="Test Author",
        )

    def setup(self) -> None:
        self.setup_called = True

    def teardown(self) -> None:
        self.teardown_called = True

    def register(self, manager) -> None:
        self.register_called = True


class FailingPlugin(Plugin[PluginConfig]):
    """Plugin that fails during setup."""

    @property
    def info(self) -> PluginInfo:
        return PluginInfo(
            name="failing-plugin",
            version="1.0.0",
            plugin_type=PluginType.CUSTOM,
        )

    def setup(self) -> None:
        raise RuntimeError("Intentional setup failure")

    def teardown(self) -> None:
        pass


# =============================================================================
# PluginInfo Tests
# =============================================================================


class TestPluginInfo:
    """Tests for PluginInfo dataclass."""

    def test_valid_plugin_info(self):
        """Test creating valid plugin info."""
        info = PluginInfo(
            name="my-plugin",
            version="1.0.0",
            plugin_type=PluginType.VALIDATOR,
            description="A test plugin",
        )
        assert info.name == "my-plugin"
        assert info.version == "1.0.0"
        assert info.plugin_type == PluginType.VALIDATOR

    def test_plugin_name_validation(self):
        """Test that invalid plugin names are rejected."""
        # Name cannot be empty
        with pytest.raises(ValueError, match="cannot be empty"):
            PluginInfo(name="", version="1.0.0", plugin_type=PluginType.CUSTOM)

        # Name must start with lowercase
        with pytest.raises(ValueError, match="Invalid plugin name"):
            PluginInfo(name="MyPlugin", version="1.0.0", plugin_type=PluginType.CUSTOM)

        # Name cannot contain spaces
        with pytest.raises(ValueError, match="Invalid plugin name"):
            PluginInfo(name="my plugin", version="1.0.0", plugin_type=PluginType.CUSTOM)

    def test_valid_plugin_names(self):
        """Test valid plugin name formats."""
        valid_names = [
            "my-plugin",
            "my_plugin",
            "myplugin",
            "plugin123",
            "a",
        ]
        for name in valid_names:
            info = PluginInfo(name=name, version="1.0.0", plugin_type=PluginType.CUSTOM)
            assert info.name == name

    def test_version_required(self):
        """Test that version is required."""
        with pytest.raises(ValueError, match="version cannot be empty"):
            PluginInfo(name="test", version="", plugin_type=PluginType.CUSTOM)

    def test_compatibility_check(self):
        """Test version compatibility checking."""
        info = PluginInfo(
            name="test",
            version="1.0.0",
            plugin_type=PluginType.CUSTOM,
            min_truthound_version="0.1.0",
            max_truthound_version="2.0.0",
        )

        # Compatible version
        assert info.is_compatible("1.0.0") is True
        assert info.is_compatible("0.5.0") is True
        assert info.is_compatible("1.9.9") is True

        # Incompatible versions
        assert info.is_compatible("0.0.1") is False
        assert info.is_compatible("2.0.1") is False
        assert info.is_compatible("3.0.0") is False

    def test_compatibility_no_constraints(self):
        """Test compatibility with no version constraints."""
        info = PluginInfo(
            name="test",
            version="1.0.0",
            plugin_type=PluginType.CUSTOM,
        )
        assert info.is_compatible("999.0.0") is True

    def test_immutability(self):
        """Test that PluginInfo is immutable."""
        info = PluginInfo(name="test", version="1.0.0", plugin_type=PluginType.CUSTOM)
        with pytest.raises(AttributeError):
            info.name = "changed"  # type: ignore


# =============================================================================
# PluginConfig Tests
# =============================================================================


class TestPluginConfig:
    """Tests for PluginConfig dataclass."""

    def test_default_config(self):
        """Test default configuration values."""
        config = PluginConfig()
        assert config.enabled is True
        assert config.priority == 100
        assert config.settings == {}
        assert config.auto_load is True

    def test_custom_config(self):
        """Test custom configuration values."""
        config = PluginConfig(
            enabled=False,
            priority=50,
            settings={"key": "value"},
            auto_load=False,
        )
        assert config.enabled is False
        assert config.priority == 50
        assert config.settings == {"key": "value"}
        assert config.auto_load is False

    def test_config_merge(self):
        """Test merging configurations."""
        base = PluginConfig(
            priority=100,
            settings={"a": 1, "b": 2},
        )
        override = PluginConfig(
            priority=50,
            settings={"b": 3, "c": 4},
        )

        merged = base.merge(override)

        assert merged.priority == 50  # Override takes precedence
        assert merged.settings == {"a": 1, "b": 3, "c": 4}  # Merged settings


# =============================================================================
# Plugin Base Class Tests
# =============================================================================


class TestPlugin:
    """Tests for Plugin base class."""

    def test_plugin_initialization(self):
        """Test plugin initialization."""
        plugin = SimpleTestPlugin()
        assert plugin.name == "simple-test"
        assert plugin.version == "1.0.0"
        assert plugin.plugin_type == PluginType.CUSTOM
        assert plugin.state == PluginState.DISCOVERED

    def test_plugin_with_config(self):
        """Test plugin with custom config."""
        config = PluginConfig(priority=50)
        plugin = SimpleTestPlugin(config)
        assert plugin.config.priority == 50

    def test_plugin_setup_teardown(self):
        """Test setup and teardown methods."""
        plugin = SimpleTestPlugin()
        assert not plugin.setup_called
        assert not plugin.teardown_called

        plugin.setup()
        assert plugin.setup_called

        plugin.teardown()
        assert plugin.teardown_called

    def test_plugin_config_change(self):
        """Test configuration change notification."""
        plugin = SimpleTestPlugin()
        plugin._state = PluginState.ACTIVE

        old_config = plugin.config
        new_config = PluginConfig(priority=25)

        # Should trigger on_config_change
        plugin.config = new_config
        assert plugin.config.priority == 25

    def test_plugin_health_check(self):
        """Test health check method."""
        plugin = SimpleTestPlugin()
        assert plugin.health_check() is False  # Not active

        plugin._state = PluginState.ACTIVE
        assert plugin.health_check() is True

    def test_plugin_repr(self):
        """Test plugin string representation."""
        plugin = SimpleTestPlugin()
        repr_str = repr(plugin)
        assert "SimpleTestPlugin" in repr_str
        assert "simple-test" in repr_str


# =============================================================================
# PluginType Tests
# =============================================================================


class TestPluginType:
    """Tests for PluginType enumeration."""

    def test_all_plugin_types(self):
        """Test all plugin types are defined."""
        expected = [
            "validator",
            "reporter",
            "datasource",
            "profiler",
            "hook",
            "transformer",
            "exporter",
            "custom",
        ]
        actual = [t.value for t in PluginType]
        assert set(expected) == set(actual)


# =============================================================================
# PluginState Tests
# =============================================================================


class TestPluginState:
    """Tests for PluginState enumeration."""

    def test_all_plugin_states(self):
        """Test all plugin states are defined."""
        expected = [
            "discovered",
            "loading",
            "loaded",
            "active",
            "inactive",
            "error",
            "unloading",
        ]
        actual = [s.value for s in PluginState]
        assert set(expected) == set(actual)


# =============================================================================
# Exception Tests
# =============================================================================


class TestPluginExceptions:
    """Tests for plugin exceptions."""

    def test_plugin_error(self):
        """Test base PluginError."""
        error = PluginError("Test error", plugin_name="test")
        assert str(error) == "Test error"
        assert error.plugin_name == "test"

    def test_plugin_load_error(self):
        """Test PluginLoadError."""
        error = PluginLoadError("Load failed", plugin_name="test")
        assert isinstance(error, PluginError)

    def test_plugin_not_found_error(self):
        """Test PluginNotFoundError."""
        error = PluginNotFoundError("Not found", plugin_name="missing")
        assert isinstance(error, PluginError)

    def test_plugin_dependency_error(self):
        """Test PluginDependencyError."""
        error = PluginDependencyError(
            "Missing deps",
            plugin_name="test",
            missing_deps=["dep1", "dep2"],
        )
        assert error.missing_deps == ["dep1", "dep2"]

    def test_plugin_compatibility_error(self):
        """Test PluginCompatibilityError."""
        error = PluginCompatibilityError(
            "Incompatible",
            plugin_name="test",
            required_version="2.0.0",
            current_version="1.0.0",
        )
        assert error.required_version == "2.0.0"
        assert error.current_version == "1.0.0"



"""Cloud DW test backend implementations.

This package contains concrete implementations of CloudDWTestBackend
for each supported cloud data warehouse platform.

Each backend provides:
    - Platform-specific connection handling
    - Credential management
    - SQL dialect translations
    - Cost estimation (where supported)
    - Dry-run capabilities (where supported)

Available Backends:
    - BigQueryTestBackend: Google BigQuery
    - SnowflakeTestBackend: Snowflake
    - RedshiftTestBackend: AWS Redshift
    - DatabricksTestBackend: Databricks SQL
    - OracleTestBackend: Oracle Database
    - SQLServerTestBackend: Microsoft SQL Server
"""

from tests.integration.cloud_dw.backends.bigquery import (
    BigQueryTestBackend,
    BigQueryCredentials,
)
from tests.integration.cloud_dw.backends.snowflake import (
    SnowflakeTestBackend,
    SnowflakeCredentials,
)
from tests.integration.cloud_dw.backends.redshift import (
    RedshiftTestBackend,
    RedshiftCredentials,
)
from tests.integration.cloud_dw.backends.databricks import (
    DatabricksTestBackend,
    DatabricksCredentials,
)
from tests.integration.cloud_dw.backends.registry import (
    BackendRegistry,
    get_backend,
    register_backend,
    get_available_backends,
)

__all__ = [
    # Backends
    "BigQueryTestBackend",
    "SnowflakeTestBackend",
    "RedshiftTestBackend",
    "DatabricksTestBackend",
    # Credentials
    "BigQueryCredentials",
    "SnowflakeCredentials",
    "RedshiftCredentials",
    "DatabricksCredentials",
    # Registry
    "BackendRegistry",
    "get_backend",
    "register_backend",
    "get_available_backends",
]



"""String length validators."""

from typing import Any

import polars as pl

from truthound.types import Severity
from truthound.validators.base import (
    ValidationIssue,
    Validator,
    StringValidatorMixin,
)
from truthound.validators.registry import register_validator


@register_validator
class LengthValidator(Validator, StringValidatorMixin):
    """Validates that string lengths are within specified range."""

    name = "length"
    category = "string"

    def __init__(
        self,
        min_length: int | None = None,
        max_length: int | None = None,
        exact_length: int | None = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.min_length = min_length
        self.max_length = max_length
        self.exact_length = exact_length

    def validate(self, lf: pl.LazyFrame) -> list[ValidationIssue]:
        issues: list[ValidationIssue] = []
        columns = self._get_string_columns(lf)

        if not columns:
            return issues

        exprs = [pl.len().alias("_total")]

        for col in columns:
            length_expr = pl.col(col).str.len_chars()

            if self.exact_length is not None:
                violation = (length_expr != self.exact_length) & pl.col(col).is_not_null()
            else:
                conditions = []
                if self.min_length is not None:
                    conditions.append(length_expr < self.min_length)
                if self.max_length is not None:
                    conditions.append(length_expr > self.max_length)

                if conditions:
                    violation = conditions[0]
                    for cond in conditions[1:]:
                        violation = violation | cond
                    violation = violation & pl.col(col).is_not_null()
                else:
                    continue

            exprs.append(violation.sum().alias(f"_inv_{col}"))

        if len(exprs) == 1:
            return issues

        result = lf.select(exprs).collect()
        total_rows = result["_total"][0]

        if total_rows == 0:
            return issues

        for col in columns:
            alias = f"_inv_{col}"
            if alias not in result.columns:
                continue

            invalid_count = result[alias][0]
            if invalid_count > 0:
                ratio = invalid_count / total_rows

                if self.exact_length is not None:
                    expected = f"exactly {self.exact_length}"
                elif self.min_length is not None and self.max_length is not None:
                    expected = f"[{self.min_length}, {self.max_length}]"
                elif self.min_length is not None:
                    expected = f">= {self.min_length}"
                else:
                    expected = f"<= {self.max_length}"

                issues.append(
                    ValidationIssue(
                        column=col,
                        issue_type="length_violation",
                        count=invalid_count,
                        severity=self._calculate_severity(ratio),
                        details=f"Length should be {expected}",
                        expected=expected,
                    )
                )

        return issues



"""Graceful degradation policies for timeout handling.

This module implements graceful degradation patterns for when validation
operations approach or exceed their time limits. Instead of failing
completely, the system can:
- Return partial results
- Use sampling to reduce workload
- Skip optional validations
- Apply fast-path algorithms

Example:
    degradation = GracefulDegradation(
        policy=DegradationPolicy(
            sample_threshold=0.3,  # Sample when 30% time remaining
            skip_threshold=0.1,    # Skip optional when 10% remaining
        )
    )

    result = degradation.apply(
        deadline_ctx,
        operation=full_validation,
        fallback=sampled_validation,
    )
"""

from __future__ import annotations

import functools
import logging
from dataclasses import dataclass, field
from datetime import datetime, timezone
from enum import Enum
from typing import Any, Callable, Generic, TypeVar

from truthound.validators.timeout.deadline import (
    DeadlineContext,
    DeadlineExceededError,
    get_current_deadline,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


class DegradationLevel(str, Enum):
    """Levels of degradation."""

    NONE = "none"           # Full quality
    LIGHT = "light"         # Minor optimizations
    MODERATE = "moderate"   # Sampling, skip optional
    HEAVY = "heavy"         # Aggressive shortcuts
    EMERGENCY = "emergency"  # Return immediately


class DegradationAction(str, Enum):
    """Actions for degradation."""

    CONTINUE = "continue"      # Continue normally
    SAMPLE = "sample"          # Use sampling
    SKIP_OPTIONAL = "skip"     # Skip optional operations
    FAST_PATH = "fast_path"    # Use fast algorithm
    CACHE_ONLY = "cache_only"  # Use cached results only
    ABORT = "abort"            # Abort operation


@dataclass
class DegradationPolicy:
    """Policy for graceful degradation.

    Attributes:
        sample_threshold: Remaining time fraction to trigger sampling
        skip_threshold: Remaining time fraction to skip optional ops
        fast_path_threshold: Remaining time fraction for fast path
        emergency_threshold: Remaining time fraction for emergency abort
        sample_ratio: Ratio of data to sample when sampling
        cache_enabled: Whether to use cached results as fallback
        notify_on_degrade: Whether to emit notifications
    """

    sample_threshold: float = 0.3       # 30% time remaining
    skip_threshold: float = 0.2         # 20% time remaining
    fast_path_threshold: float = 0.1    # 10% time remaining
    emergency_threshold: float = 0.05   # 5% time remaining
    sample_ratio: float = 0.1           # Sample 10% of data
    cache_enabled: bool = True
    notify_on_degrade: bool = True

    def get_degradation_level(
        self,
        remaining_fraction: float,
    ) -> DegradationLevel:
        """Get degradation level based on remaining time.

        Args:
            remaining_fraction: Fraction of time remaining (0.0-1.0)

        Returns:
            Appropriate DegradationLevel
        """
        if remaining_fraction <= self.emergency_threshold:
            return DegradationLevel.EMERGENCY
        elif remaining_fraction <= self.fast_path_threshold:
            return DegradationLevel.HEAVY
        elif remaining_fraction <= self.skip_threshold:
            return DegradationLevel.MODERATE
        elif remaining_fraction <= self.sample_threshold:
            return DegradationLevel.LIGHT
        return DegradationLevel.NONE

    def get_action(self, level: DegradationLevel) -> DegradationAction:
        """Get action for degradation level.

        Args:
            level: Current degradation level

        Returns:
            Recommended action
        """
        actions = {
            DegradationLevel.NONE: DegradationAction.CONTINUE,
            DegradationLevel.LIGHT: DegradationAction.SAMPLE,
            DegradationLevel.MODERATE: DegradationAction.SKIP_OPTIONAL,
            DegradationLevel.HEAVY: DegradationAction.FAST_PATH,
            DegradationLevel.EMERGENCY: DegradationAction.ABORT,
        }
        return actions.get(level, DegradationAction.CONTINUE)


@dataclass
class DegradationResult(Generic[T]):
    """Result of a degraded operation.

    Attributes:
        value: Result value
        level: Degradation level applied
        action: Action taken
        quality_score: Estimated quality (0.0-1.0)
        was_degraded: Whether degradation was applied
        fallback_used: Whether fallback was used
        metadata: Additional metadata
    """

    value: T | None
    level: DegradationLevel = DegradationLevel.NONE
    action: DegradationAction = DegradationAction.CONTINUE
    quality_score: float = 1.0
    was_degraded: bool = False
    fallback_used: bool = False
    error: str | None = None
    metadata: dict[str, Any] = field(default_factory=dict)
    timestamp: datetime = field(default_factory=lambda: datetime.now(timezone.utc))

    @classmethod
    def success(cls, value: T, level: DegradationLevel = DegradationLevel.NONE) -> "DegradationResult[T]":
        """Create successful result.

        Args:
            value: Result value
            level: Degradation level applied

        Returns:
            Success result
        """
        quality = {
            DegradationLevel.NONE: 1.0,
            DegradationLevel.LIGHT: 0.9,
            DegradationLevel.MODERATE: 0.7,
            DegradationLevel.HEAVY: 0.5,
            DegradationLevel.EMERGENCY: 0.2,
        }
        return cls(
            value=value,
            level=level,
            action=DegradationAction.CONTINUE,
            quality_score=quality.get(level, 1.0),
            was_degraded=level != DegradationLevel.NONE,
        )

    @classmethod
    def fallback(
        cls,
        value: T | None,
        level: DegradationLevel,
        action: DegradationAction,
    ) -> "DegradationResult[T]":
        """Create fallback result.

        Args:
            value: Fallback value
            level: Degradation level
            action: Action taken

        Returns:
            Fallback result
        """
        quality = {
            DegradationLevel.LIGHT: 0.8,
            DegradationLevel.MODERATE: 0.6,
            DegradationLevel.HEAVY: 0.4,
            DegradationLevel.EMERGENCY: 0.1,
        }
        return cls(
            value=value,
            level=level,
            action=action,
            quality_score=quality.get(level, 0.5),
            was_degraded=True,
            fallback_used=True,
        )

    @classmethod
    def aborted(cls, error: str | None = None) -> "DegradationResult[T]":
        """Create aborted result.

        Args:
            error: Error message

        Returns:
            Aborted result
        """
        return cls(
            value=None,
            level=DegradationLevel.EMERGENCY,
            action=DegradationAction.ABORT,
            quality_score=0.0,
            was_degraded=True,
            error=error,
        )

    def to_dict(self) -> dict[str, Any]:
        """Convert to dictionary."""
        return {
            "level": self.level.value,
            "action": self.action.value,
            "quality_score": self.quality_score,
            "was_degraded": self.was_degraded,
            "fallback_used": self.fallback_used,
            "error": self.error,
            "metadata": self.metadata,
            "timestamp": self.timestamp.isoformat(),
        }


@dataclass
class GracefulDegradation:
    """Manager for graceful degradation.

    This class manages the application of degradation policies to
    operations that may approach their time limits.

    Example:
        degradation = GracefulDegradation()

        # Check if degradation needed
        level = degradation.check(deadline_ctx)

        if level != DegradationLevel.NONE:
            # Use degraded operation
            result = fast_validate(data)
        else:
            # Full validation
            result = full_validate(data)

        # Or use apply() for automatic handling
        result = degradation.apply(
            deadline_ctx,
            full_operation=full_validate,
            fallback_operation=fast_validate,
        )
    """

    policy: DegradationPolicy = field(default_factory=DegradationPolicy)
    _cache: dict[str, Any] = field(default_factory=dict, init=False)
    _stats: dict[str, int] = field(default_factory=dict, init=False)

    def check(
        self,
        context: DeadlineContext | None = None,
    ) -> DegradationLevel:
        """Check current degradation level.

        Args:
            context: Deadline context (uses current if None)

        Returns:
            Current DegradationLevel
        """
        if context is None:
            context = get_current_deadline()

        if context is None:
            return DegradationLevel.NONE

        # Calculate remaining fraction
        elapsed = (
            datetime.now(timezone.utc) - context.created_at
        ).total_seconds()
        total = (context.deadline_utc - context.created_at).total_seconds()

        if total <= 0:
            return DegradationLevel.EMERGENCY

        remaining_fraction = context.remaining_seconds / total

        return self.policy.get_degradation_level(remaining_fraction)

    def get_action(
        self,
        context: DeadlineContext | None = None,
    ) -> DegradationAction:
        """Get recommended action based on current state.

        Args:
            context: Deadline context

        Returns:
            Recommended DegradationAction
        """
        level = self.check(context)
        return self.policy.get_action(level)

    def should_sample(self, context: DeadlineContext | None = None) -> bool:
        """Check if sampling should be used.

        Args:
            context: Deadline context

        Returns:
            True if sampling recommended
        """
        level = self.check(context)
        return level in (
            DegradationLevel.LIGHT,
            DegradationLevel.MODERATE,
        )

    def should_skip_optional(self, context: DeadlineContext | None = None) -> bool:
        """Check if optional operations should be skipped.

        Args:
            context: Deadline context

        Returns:
            True if skipping recommended
        """
        level = self.check(context)
        return level in (
            DegradationLevel.MODERATE,
            DegradationLevel.HEAVY,
            DegradationLevel.EMERGENCY,
        )

    def should_abort(self, context: DeadlineContext | None = None) -> bool:
        """Check if operation should abort.

        Args:
            context: Deadline context

        Returns:
            True if abort recommended
        """
        level = self.check(context)
        return level == DegradationLevel.EMERGENCY

    def apply(
        self,
        context: DeadlineContext | None,
        full_operation: Callable[[], T],
        fallback_operation: Callable[[], T] | None = None,
        cache_key: str | None = None,
        operation_name: str = "",
    ) -> DegradationResult[T]:
        """Apply degradation policy to an operation.

        Args:
            context: Deadline context
            full_operation: Full quality operation
            fallback_operation: Fallback operation for degradation
            cache_key: Key for caching results
            operation_name: Name for logging

        Returns:
            DegradationResult with value and metadata
        """
        level = self.check(context)
        action = self.policy.get_action(level)

        # Track statistics
        self._track_stat(f"level_{level.value}")
        self._track_stat(f"action_{action.value}")

        # Handle emergency abort
        if action == DegradationAction.ABORT:
            logger.warning(f"Emergency abort for '{operation_name}'")
            return DegradationResult.aborted(
                f"Emergency abort due to time pressure"
            )

        # Try cache if available
        if cache_key and self.policy.cache_enabled and action != DegradationAction.CONTINUE:
            cached = self._cache.get(cache_key)
            if cached is not None:
                logger.debug(f"Using cached result for '{operation_name}'")
                return DegradationResult.fallback(
                    cached,
                    level,
                    DegradationAction.CACHE_ONLY,
                )

        # Use fallback if degraded
        if level != DegradationLevel.NONE and fallback_operation is not None:
            try:
                result = fallback_operation()
                if cache_key:
                    self._cache[cache_key] = result
                return DegradationResult.fallback(result, level, action)
            except Exception as e:
                logger.warning(f"Fallback failed: {e}")
                # Fall through to try full operation

        # Try full operation
        try:
            result = full_operation()
            if cache_key:
                self._cache[cache_key] = result
            return DegradationResult.success(result, level)
        except DeadlineExceededError as e:
            logger.warning(f"Deadline exceeded in '{operation_name}': {e}")
            return DegradationResult.aborted(str(e))
        except TimeoutError as e:
            logger.warning(f"Timeout in '{operation_name}': {e}")
            return DegradationResult.aborted(str(e))
        except Exception as e:
            logger.error(f"Error in '{operation_name}': {e}")
            raise

    def _track_stat(self, key: str) -> None:
        """Track a statistic.

        Args:
            key: Statistic key
        """
        self._stats[key] = self._stats.get(key, 0) + 1

    def get_stats(self) -> dict[str, int]:
        """Get degradation statistics.

        Returns:
            Dictionary of statistics
        """
        return dict(self._stats)

    def clear_cache(self) -> None:
        """Clear the result cache."""
        self._cache.clear()

    def clear_stats(self) -> None:
        """Clear statistics."""
        self._stats.clear()


def with_graceful_degradation(
    fallback: Callable[..., T] | None = None,
    policy: DegradationPolicy | None = None,
) -> Callable[[Callable[..., T]], Callable[..., DegradationResult[T]]]:
    """Decorator to add graceful degradation to a function.

    Args:
        fallback: Fallback function to use when degraded
        policy: Degradation policy

    Returns:
        Decorated function

    Example:
        @with_graceful_degradation(fallback=fast_validate)
        def full_validate(data):
            ...
    """
    degradation = GracefulDegradation(policy or DegradationPolicy())

    def decorator(func: Callable[..., T]) -> Callable[..., DegradationResult[T]]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> DegradationResult[T]:
            ctx = get_current_deadline()

            full_op = lambda: func(*args, **kwargs)
            fallback_op = (
                (lambda: fallback(*args, **kwargs)) if fallback else None
            )

            return degradation.apply(
                context=ctx,
                full_operation=full_op,
                fallback_operation=fallback_op,
                operation_name=func.__name__,
            )

        return wrapper

    return decorator


# Default degradation instance
_default_degradation = GracefulDegradation()


def check_degradation_level() -> DegradationLevel:
    """Check current degradation level using default instance.

    Returns:
        Current DegradationLevel
    """
    return _default_degradation.check()


def should_degrade() -> bool:
    """Check if any degradation should be applied.

    Returns:
        True if degradation recommended
    """
    level = check_degradation_level()
    return level != DegradationLevel.NONE



"""Tests for the enterprise connection pool manager.

This module provides comprehensive tests for:
- Connection pool configuration
- Pool strategies
- Circuit breaker
- Retry handler
- Health checks
- Metrics collection
- Database dialect detection
"""

from __future__ import annotations

import threading
import time
from datetime import datetime, timedelta
from typing import Any
from unittest.mock import MagicMock, patch, PropertyMock

import pytest

# Check if sqlalchemy is available
try:
    import sqlalchemy
    from sqlalchemy import text
    from sqlalchemy.exc import OperationalError, DisconnectionError

    HAS_SQLALCHEMY = True
except ImportError:
    HAS_SQLALCHEMY = False
    sqlalchemy = None  # type: ignore

# Import the module under test
from truthound.stores.backends.connection_pool import (
    CircuitBreaker,
    CircuitBreakerConfig,
    CircuitState,
    ConnectionPoolConfig,
    ConnectionPoolManager,
    DatabaseDialect,
    HealthCheckConfig,
    PoolConfig,
    PoolMetrics,
    PoolStrategy,
    RetryConfig,
    RetryHandler,
    create_pool_manager,
    create_pool_for_dialect,
)


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture
def pool_config() -> PoolConfig:
    """Create a default pool configuration."""
    return PoolConfig(
        strategy=PoolStrategy.QUEUE_POOL,
        pool_size=5,
        max_overflow=10,
        pool_timeout=30.0,
        pool_recycle=3600,
        pool_pre_ping=True,
    )


@pytest.fixture
def retry_config() -> RetryConfig:
    """Create a default retry configuration."""
    return RetryConfig(
        max_retries=3,
        base_delay=0.01,  # Fast for tests
        max_delay=0.1,
        exponential_base=2.0,
        jitter=False,  # Deterministic for tests
    )


@pytest.fixture
def circuit_breaker_config() -> CircuitBreakerConfig:
    """Create a default circuit breaker configuration."""
    return CircuitBreakerConfig(
        failure_threshold=3,
        success_threshold=2,
        timeout=1.0,  # Fast for tests
        half_open_max_calls=2,
    )


@pytest.fixture
def health_check_config() -> HealthCheckConfig:
    """Create a default health check configuration."""
    return HealthCheckConfig(
        enabled=False,  # Disabled for tests
        interval=1.0,
        timeout=1.0,
        query="SELECT 1",
    )


@pytest.fixture
def connection_pool_config(
    pool_config: PoolConfig,
    retry_config: RetryConfig,
    circuit_breaker_config: CircuitBreakerConfig,
    health_check_config: HealthCheckConfig,
) -> ConnectionPoolConfig:
    """Create a complete connection pool configuration."""
    return ConnectionPoolConfig(
        connection_url="sqlite:///:memory:",
        pool=pool_config,
        retry=retry_config,
        circuit_breaker=circuit_breaker_config,
        health_check=health_check_config,
        echo=False,
    )


# =============================================================================
# DatabaseDialect Tests
# =============================================================================


class TestDatabaseDialect:
    """Tests for database dialect detection."""

    @pytest.mark.parametrize(
        "url,expected",
        [
            ("postgresql://user:pass@localhost/db", DatabaseDialect.POSTGRESQL),
            ("postgres://user:pass@localhost/db", DatabaseDialect.POSTGRESQL),
            ("postgresql+psycopg2://user:pass@localhost/db", DatabaseDialect.POSTGRESQL),
            ("mysql://user:pass@localhost/db", DatabaseDialect.MYSQL),
            ("mysql+pymysql://user:pass@localhost/db", DatabaseDialect.MYSQL),
            ("sqlite:///path/to/db.sqlite", DatabaseDialect.SQLITE),
            ("sqlite:///:memory:", DatabaseDialect.SQLITE),
            ("mssql+pyodbc://user:pass@server/db", DatabaseDialect.MSSQL),
            ("oracle://user:pass@localhost/db", DatabaseDialect.ORACLE),
            ("unknown://localhost/db", DatabaseDialect.UNKNOWN),
        ],
    )
    def test_dialect_from_url(self, url: str, expected: DatabaseDialect) -> None:
        """Test dialect detection from various URLs."""
        assert DatabaseDialect.from_url(url) == expected


# =============================================================================
# PoolConfig Tests
# =============================================================================


class TestPoolConfig:
    """Tests for pool configuration."""

    def test_default_values(self) -> None:
        """Test default configuration values."""
        config = PoolConfig()
        assert config.strategy == PoolStrategy.QUEUE_POOL
        assert config.pool_size == 5
        assert config.max_overflow == 10
        assert config.pool_timeout == 30.0
        assert config.pool_recycle == 3600
        assert config.pool_pre_ping is True

    def test_custom_values(self) -> None:
        """Test custom configuration values."""
        config = PoolConfig(
            strategy=PoolStrategy.NULL_POOL,
            pool_size=20,
            max_overflow=30,
            pool_timeout=60.0,
            pool_recycle=1800,
            pool_pre_ping=False,
        )
        assert config.strategy == PoolStrategy.NULL_POOL
        assert config.pool_size == 20
        assert config.max_overflow == 30


# =============================================================================
# ConnectionPoolConfig Tests
# =============================================================================


class TestConnectionPoolConfig:
    """Tests for complete connection pool configuration."""

    def test_dialect_property(self) -> None:
        """Test dialect property detection."""
        config = ConnectionPoolConfig(
            connection_url="postgresql://localhost/db"
        )
        assert config.dialect == DatabaseDialect.POSTGRESQL

    def test_with_dialect_defaults_sqlite(self) -> None:
        """Test SQLite-specific defaults."""
        config = ConnectionPoolConfig(
            connection_url="sqlite:///:memory:"
        )
        updated = config.with_dialect_defaults()
        assert updated.connect_args.get("check_same_thread") is False
        # SQLite should have reduced pool size
        assert updated.pool.pool_size == 1
        assert updated.pool.max_overflow == 0

    def test_with_dialect_defaults_postgresql(self) -> None:
        """Test PostgreSQL-specific defaults."""
        config = ConnectionPoolConfig(
            connection_url="postgresql://localhost/db"
        )
        updated = config.with_dialect_defaults()
        assert updated.engine_options.get("pool_use_lifo") is True
        assert updated.pool.pool_recycle <= 1800

    def test_with_dialect_defaults_mysql(self) -> None:
        """Test MySQL-specific defaults."""
        config = ConnectionPoolConfig(
            connection_url="mysql://localhost/db"
        )
        updated = config.with_dialect_defaults()
        assert updated.connect_args.get("connect_timeout") == 10


# =============================================================================
# PoolMetrics Tests
# =============================================================================


class TestPoolMetrics:
    """Tests for pool metrics collection."""

    def test_initial_values(self) -> None:
        """Test initial metric values."""
        metrics = PoolMetrics()
        assert metrics.total_connections == 0
        assert metrics.active_connections == 0
        assert metrics.idle_connections == 0
        assert metrics.checkouts == 0
        assert metrics.checkins == 0

    def test_record_checkout(self) -> None:
        """Test checkout recording."""
        metrics = PoolMetrics()
        metrics.record_checkout(10.5)
        assert metrics.checkouts == 1
        assert metrics.current_checked_out == 1
        assert metrics.total_checkout_time_ms == 10.5
        assert metrics.peak_connections == 1

    def test_record_checkin(self) -> None:
        """Test checkin recording."""
        metrics = PoolMetrics()
        metrics.record_checkout(5.0)
        metrics.record_checkin()
        assert metrics.checkins == 1
        assert metrics.current_checked_out == 0

    def test_record_creation(self) -> None:
        """Test connection creation recording."""
        metrics = PoolMetrics()
        metrics.record_creation()
        assert metrics.total_created == 1
        assert metrics.current_size == 1

    def test_record_close(self) -> None:
        """Test connection close recording."""
        metrics = PoolMetrics()
        metrics.record_creation()
        metrics.record_close()
        assert metrics.total_closed == 1
        assert metrics.current_size == 0

    def test_record_health_check(self) -> None:
        """Test health check recording."""
        metrics = PoolMetrics()
        metrics.record_health_check(True)
        assert metrics.health_checks_passed == 1
        assert metrics.last_health_check is not None

        metrics.record_health_check(False)
        assert metrics.health_checks_failed == 1

    def test_avg_checkout_time(self) -> None:
        """Test average checkout time calculation."""
        metrics = PoolMetrics()
        metrics.record_checkout(10.0)
        metrics.record_checkout(20.0)
        metrics.record_checkout(30.0)
        assert metrics.avg_checkout_time_ms == 20.0

    def test_to_dict(self) -> None:
        """Test metrics serialization to dict."""
        metrics = PoolMetrics()
        metrics.record_checkout(10.0)
        metrics.record_creation()

        data = metrics.to_dict()
        assert "connections" in data
        assert "operations" in data
        assert "timing" in data
        assert "health" in data
        assert "circuit_breaker" in data

    def test_to_prometheus(self) -> None:
        """Test Prometheus format export."""
        metrics = PoolMetrics()
        metrics.record_checkout(10.0)

        prometheus = metrics.to_prometheus()
        assert "truthound_pool_checkouts_total 1" in prometheus
        assert "truthound_pool_connections_active 1" in prometheus

    def test_thread_safety(self) -> None:
        """Test thread-safe metric updates."""
        metrics = PoolMetrics()
        threads = []

        def update_metrics() -> None:
            for _ in range(100):
                metrics.record_checkout(1.0)
                metrics.record_checkin()

        for _ in range(10):
            t = threading.Thread(target=update_metrics)
            threads.append(t)
            t.start()

        for t in threads:
            t.join()

        assert metrics.checkouts == 1000
        assert metrics.checkins == 1000


# =============================================================================
# CircuitBreaker Tests
# =============================================================================


class TestCircuitBreaker:
    """Tests for circuit breaker pattern."""

    def test_initial_state(self, circuit_breaker_config: CircuitBreakerConfig) -> None:
        """Test initial circuit state is closed."""
        cb = CircuitBreaker(circuit_breaker_config)
        assert cb.state == CircuitState.CLOSED
        assert cb.is_closed
        assert not cb.is_open
        assert cb.can_execute()

    def test_opens_after_threshold(
        self, circuit_breaker_config: CircuitBreakerConfig
    ) -> None:
        """Test circuit opens after failure threshold."""
        cb = CircuitBreaker(circuit_breaker_config)

        for _ in range(circuit_breaker_config.failure_threshold):
            cb.record_failure()

        assert cb.state == CircuitState.OPEN
        assert cb.is_open
        assert not cb.can_execute()

    def test_transitions_to_half_open(
        self, circuit_breaker_config: CircuitBreakerConfig
    ) -> None:
        """Test circuit transitions to half-open after timeout."""
        cb = CircuitBreaker(circuit_breaker_config)

        # Open the circuit
        for _ in range(circuit_breaker_config.failure_threshold):
            cb.record_failure()

        assert cb.is_open

        # Wait for timeout
        time.sleep(circuit_breaker_config.timeout + 0.1)

        # Should transition to half-open on next check
        assert cb.can_execute()
        assert cb.state == CircuitState.HALF_OPEN

    def test_closes_after_success_threshold(
        self, circuit_breaker_config: CircuitBreakerConfig
    ) -> None:
        """Test circuit closes after success threshold in half-open."""
        cb = CircuitBreaker(circuit_breaker_config)

        # Open the circuit
        for _ in range(circuit_breaker_config.failure_threshold):
            cb.record_failure()

        # Wait for timeout
        time.sleep(circuit_breaker_config.timeout + 0.1)

        # Transition to half-open
        cb.can_execute()
        assert cb.state == CircuitState.HALF_OPEN

        # Record successes
        for _ in range(circuit_breaker_config.success_threshold):
            cb.record_success()

        assert cb.state == CircuitState.CLOSED
        assert cb.is_closed

    def test_half_open_failure_reopens(
        self, circuit_breaker_config: CircuitBreakerConfig
    ) -> None:
        """Test failure in half-open immediately reopens circuit."""
        cb = CircuitBreaker(circuit_breaker_config)

        # Open the circuit
        for _ in range(circuit_breaker_config.failure_threshold):
            cb.record_failure()

        # Wait for timeout
        time.sleep(circuit_breaker_config.timeout + 0.1)

        # Transition to half-open
        cb.can_execute()
        cb.record_failure()

        assert cb.state == CircuitState.OPEN

    def test_reset(self, circuit_breaker_config: CircuitBreakerConfig) -> None:
        """Test manual circuit reset."""
        cb = CircuitBreaker(circuit_breaker_config)

        # Open the circuit
        for _ in range(circuit_breaker_config.failure_threshold):
            cb.record_failure()

        assert cb.is_open

        cb.reset()
        assert cb.is_closed
        assert cb.can_execute()

    def test_half_open_max_calls(
        self, circuit_breaker_config: CircuitBreakerConfig
    ) -> None:
        """Test half-open limits concurrent calls."""
        cb = CircuitBreaker(circuit_breaker_config)

        # Open and transition to half-open
        for _ in range(circuit_breaker_config.failure_threshold):
            cb.record_failure()

        time.sleep(circuit_breaker_config.timeout + 0.1)

        # Use all half-open calls
        for _ in range(circuit_breaker_config.half_open_max_calls):
            assert cb.can_execute()

        # Next call should be rejected
        assert not cb.can_execute()


# =============================================================================
# RetryHandler Tests
# =============================================================================


class TestRetryHandler:
    """Tests for retry handler."""

    def test_should_retry_on_retryable_error(
        self, retry_config: RetryConfig
    ) -> None:
        """Test retry decision for retryable errors."""
        handler = RetryHandler(retry_config)

        if HAS_SQLALCHEMY:
            error = OperationalError("test", None, None)
            assert handler.should_retry(error, 1)
            assert handler.should_retry(error, 2)
            assert not handler.should_retry(error, 3)  # Max retries

    def test_should_not_retry_on_other_errors(
        self, retry_config: RetryConfig
    ) -> None:
        """Test no retry for non-retryable errors."""
        handler = RetryHandler(retry_config)
        error = ValueError("test")
        assert not handler.should_retry(error, 1)

    def test_get_delay_exponential(self, retry_config: RetryConfig) -> None:
        """Test exponential backoff delay calculation."""
        handler = RetryHandler(retry_config)

        delay1 = handler.get_delay(1)
        delay2 = handler.get_delay(2)
        delay3 = handler.get_delay(3)

        assert delay1 == pytest.approx(0.01, rel=0.1)
        assert delay2 == pytest.approx(0.02, rel=0.1)
        assert delay3 == pytest.approx(0.04, rel=0.1)

    def test_get_delay_max_cap(self) -> None:
        """Test delay is capped at max_delay."""
        config = RetryConfig(
            max_retries=10,
            base_delay=1.0,
            max_delay=5.0,
            exponential_base=10.0,
            jitter=False,
        )
        handler = RetryHandler(config)

        # Very high attempt should be capped
        delay = handler.get_delay(10)
        assert delay <= 5.0

    def test_execute_with_retry_success(
        self, retry_config: RetryConfig
    ) -> None:
        """Test successful execution without retry."""
        handler = RetryHandler(retry_config)
        call_count = 0

        def operation() -> str:
            nonlocal call_count
            call_count += 1
            return "success"

        result = handler.execute_with_retry(operation)
        assert result == "success"
        assert call_count == 1

    def test_execute_with_retry_eventual_success(
        self, retry_config: RetryConfig
    ) -> None:
        """Test eventual success after retries."""
        if not HAS_SQLALCHEMY:
            pytest.skip("SQLAlchemy not available")

        handler = RetryHandler(retry_config)
        call_count = 0

        def operation() -> str:
            nonlocal call_count
            call_count += 1
            if call_count < 3:
                raise OperationalError("test", None, None)
            return "success"

        result = handler.execute_with_retry(operation)
        assert result == "success"
        assert call_count == 3

    def test_execute_with_retry_exhausted(
        self, retry_config: RetryConfig
    ) -> None:
        """Test exception raised when retries exhausted."""
        if not HAS_SQLALCHEMY:
            pytest.skip("SQLAlchemy not available")

        handler = RetryHandler(retry_config)

        def operation() -> str:
            raise OperationalError("test", None, None)

        with pytest.raises(OperationalError):
            handler.execute_with_retry(operation)


# =============================================================================
# ConnectionPoolManager Tests
# =============================================================================


@pytest.mark.skipif(not HAS_SQLALCHEMY, reason="SQLAlchemy not available")
class TestConnectionPoolManager:
    """Tests for connection pool manager."""

    def test_initialization(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test pool manager initialization."""
        manager = ConnectionPoolManager(connection_pool_config)
        manager.initialize()

        assert manager.is_healthy
        assert manager.circuit_state == CircuitState.CLOSED

        manager.dispose()

    def test_context_manager(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test context manager protocol."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            assert manager.is_healthy

    def test_get_engine(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test getting the SQLAlchemy engine."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            engine = manager.get_engine()
            assert engine is not None

    def test_get_session(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test getting a database session."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            session = manager.get_session()
            assert session is not None
            session.close()

    def test_session_context_manager(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test session context manager."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            with manager.session() as session:
                result = session.execute(text("SELECT 1"))
                assert result.scalar() == 1

    def test_session_auto_commit(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test session auto-commits on success."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            with manager.session() as session:
                session.execute(
                    text("CREATE TABLE test_table (id INTEGER PRIMARY KEY)")
                )

            # Table should exist after commit
            with manager.session() as session:
                result = session.execute(
                    text("SELECT name FROM sqlite_master WHERE type='table'")
                )
                tables = [row[0] for row in result]
                assert "test_table" in tables

    def test_session_auto_rollback(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test session auto-rollback on error."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            # Create table first
            with manager.session() as session:
                session.execute(
                    text("CREATE TABLE rollback_test (id INTEGER PRIMARY KEY)")
                )

            # Try insert with error
            try:
                with manager.session() as session:
                    session.execute(
                        text("INSERT INTO rollback_test VALUES (1)")
                    )
                    raise ValueError("Simulated error")
            except ValueError:
                pass

            # Insert should have been rolled back
            with manager.session() as session:
                result = session.execute(
                    text("SELECT COUNT(*) FROM rollback_test")
                )
                assert result.scalar() == 0

    def test_metrics_collection(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test metrics are collected during operations."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            with manager.session() as session:
                session.execute(text("SELECT 1"))

            metrics = manager.metrics
            assert metrics.checkouts >= 1

    def test_pool_status(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test pool status reporting."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            status = manager.get_pool_status()

            assert status["initialized"] is True
            assert status["healthy"] is True
            assert status["circuit_state"] == "CLOSED"
            assert "config" in status
            assert "metrics" in status

    def test_password_masking(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test password is masked in status output."""
        config = ConnectionPoolConfig(
            connection_url="postgresql://user:secretpass@localhost/db",
            health_check=HealthCheckConfig(enabled=False),
        )

        # Just test the masking function directly
        manager = ConnectionPoolManager.__new__(ConnectionPoolManager)
        masked = manager._mask_password(config.connection_url)
        assert "secretpass" not in masked
        assert "***" in masked

    def test_recycle_connections(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test manual connection recycling."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            # Use a connection first
            with manager.session() as session:
                session.execute(text("SELECT 1"))

            # Recycle
            recycled = manager.recycle_connections()
            # SQLite may not track this well, just verify it runs
            assert isinstance(recycled, int)

    def test_circuit_breaker_integration(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test circuit breaker integration."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            # Record failures to open circuit
            for _ in range(
                connection_pool_config.circuit_breaker.failure_threshold
            ):
                manager._circuit_breaker.record_failure()

            assert manager.circuit_state == CircuitState.OPEN

            # Session should be rejected
            with pytest.raises(RuntimeError, match="Circuit breaker is open"):
                manager.get_session()

    def test_factory_simple_params(self) -> None:
        """Test factory with simple parameters."""
        manager = create_pool_manager(
            "sqlite:///:memory:",
            pool_size=3,
            max_overflow=5,
        )
        manager.initialize()
        assert manager.is_healthy
        manager.dispose()

    def test_factory_with_options(self) -> None:
        """Test factory with additional options."""
        manager = create_pool_manager(
            "sqlite:///:memory:",
            pool_size=10,
            max_overflow=20,
            strategy=PoolStrategy.STATIC_POOL,
            enable_circuit_breaker=False,
            enable_health_checks=False,
        )
        manager.initialize()
        assert manager.is_healthy
        manager.dispose()


@pytest.mark.skipif(not HAS_SQLALCHEMY, reason="SQLAlchemy not available")
class TestCreatePoolForDialect:
    """Tests for dialect-specific pool creation."""

    def test_sqlite_dialect(self) -> None:
        """Test SQLite pool creation."""
        manager = create_pool_for_dialect(
            DatabaseDialect.SQLITE,
            database=":memory:",
        )
        manager.initialize()
        assert manager.is_healthy
        manager.dispose()

    def test_postgresql_url_format(self) -> None:
        """Test PostgreSQL URL format."""
        # Just test URL construction, not actual connection
        manager = create_pool_for_dialect.__wrapped__  # type: ignore
        # We can't test actual connection without PostgreSQL

    def test_invalid_dialect(self) -> None:
        """Test error on invalid dialect."""
        with pytest.raises(ValueError):
            create_pool_for_dialect("invalid_dialect")  # type: ignore


# =============================================================================
# Concurrency Tests
# =============================================================================


@pytest.mark.skipif(not HAS_SQLALCHEMY, reason="SQLAlchemy not available")
class TestConcurrency:
    """Tests for concurrent pool operations."""

    def test_concurrent_sessions(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test concurrent session access."""
        # Increase pool size for concurrency test
        connection_pool_config.pool.pool_size = 5
        connection_pool_config.pool.max_overflow = 10

        with ConnectionPoolManager(connection_pool_config) as manager:
            results: list[int] = []
            errors: list[Exception] = []
            lock = threading.Lock()

            def worker(worker_id: int) -> None:
                try:
                    with manager.session() as session:
                        result = session.execute(text(f"SELECT {worker_id}"))
                        value = result.scalar()
                        with lock:
                            results.append(value)
                except Exception as e:
                    with lock:
                        errors.append(e)

            threads = [
                threading.Thread(target=worker, args=(i,)) for i in range(10)
            ]

            for t in threads:
                t.start()

            for t in threads:
                t.join()

            assert len(errors) == 0
            assert len(results) == 10
            assert sorted(results) == list(range(10))

    def test_concurrent_metric_updates(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test thread-safe metric updates under concurrency."""
        with ConnectionPoolManager(connection_pool_config) as manager:
            def worker() -> None:
                for _ in range(50):
                    with manager.session() as session:
                        session.execute(text("SELECT 1"))

            threads = [threading.Thread(target=worker) for _ in range(5)]

            for t in threads:
                t.start()

            for t in threads:
                t.join()

            # All operations should be tracked
            assert manager.metrics.checkouts >= 250


# =============================================================================
# Edge Cases
# =============================================================================


@pytest.mark.skipif(not HAS_SQLALCHEMY, reason="SQLAlchemy not available")
class TestEdgeCases:
    """Tests for edge cases and error handling."""

    def test_dispose_before_init(self) -> None:
        """Test disposing before initialization."""
        config = ConnectionPoolConfig(
            connection_url="sqlite:///:memory:",
            health_check=HealthCheckConfig(enabled=False),
        )
        manager = ConnectionPoolManager(config)
        manager.dispose()  # Should not raise

    def test_double_dispose(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test double dispose is safe."""
        manager = ConnectionPoolManager(connection_pool_config)
        manager.initialize()
        manager.dispose()
        manager.dispose()  # Should not raise

    def test_init_after_dispose(
        self, connection_pool_config: ConnectionPoolConfig
    ) -> None:
        """Test initialization after dispose raises error."""
        manager = ConnectionPoolManager(connection_pool_config)
        manager.initialize()
        manager.dispose()

        with pytest.raises(RuntimeError, match="Cannot initialize disposed"):
            manager.initialize()

    def test_empty_metrics(self) -> None:
        """Test metrics with no operations."""
        metrics = PoolMetrics()
        assert metrics.avg_checkout_time_ms == 0.0
        assert metrics.total_connections == 0

    def test_negative_metrics_protection(self) -> None:
        """Test metrics don't go negative."""
        metrics = PoolMetrics()
        metrics.record_checkin()  # Without checkout
        assert metrics.current_checked_out == 0

        metrics.record_close()  # Without creation
        assert metrics.current_size == 0



"""Task detail view.

Renders task metrics for display in various formats.
"""

from __future__ import annotations

from datetime import datetime
from typing import Any

from truthound.checkpoint.monitoring.protocols import (
    QueueMetrics,
    WorkerMetrics,
    TaskMetrics,
)
from truthound.checkpoint.monitoring.views.base import BaseView


class TaskDetailView(BaseView):
    """View for rendering task details.

    Example:
        >>> view = TaskDetailView()
        >>>
        >>> # Render single task
        >>> output = view.render(task_metrics)
        >>>
        >>> # Render for CLI
        >>> print(view.format_for_cli(task_metrics))
        Task: task-abc123
        ├── Checkpoint: daily_validation
        ├── State: running [>]
        ├── Queue: default
        ├── Worker: worker-1
        ├── Wait Time: 150ms
        └── Running For: 2.5s
    """

    def __init__(self, name: str = "task_detail") -> None:
        """Initialize task detail view."""
        super().__init__(name=name)

    def render(
        self,
        metrics: QueueMetrics | WorkerMetrics | TaskMetrics,
    ) -> dict[str, Any]:
        """Render task metrics as dictionary."""
        if not isinstance(metrics, TaskMetrics):
            return {"error": "Expected TaskMetrics"}

        return {
            "task_id": metrics.task_id,
            "checkpoint_name": metrics.checkpoint_name,
            "state": metrics.state,
            "queue_name": metrics.queue_name,
            "worker_id": metrics.worker_id,
            "timing": {
                "submitted_at": metrics.submitted_at.isoformat(),
                "started_at": metrics.started_at.isoformat() if metrics.started_at else None,
                "completed_at": metrics.completed_at.isoformat() if metrics.completed_at else None,
                "wait_time_ms": metrics.wait_time_ms,
                "execution_time_ms": metrics.execution_time_ms,
                "total_time_ms": metrics.total_time_ms,
            },
            "retries": metrics.retries,
            "error": metrics.error,
        }

    def render_summary(
        self,
        queue_metrics: list[QueueMetrics],
        worker_metrics: list[WorkerMetrics],
    ) -> dict[str, Any]:
        """Render task summary (placeholder - tasks are typically listed individually)."""
        return {
            "note": "Task view is typically used for individual tasks",
            "queues": len(queue_metrics),
            "workers": len(worker_metrics),
        }

    def render_task_list(
        self,
        tasks: list[TaskMetrics],
    ) -> dict[str, Any]:
        """Render a list of tasks."""
        by_state: dict[str, list[dict]] = {}

        for task in tasks:
            rendered = self.render(task)
            state = task.state
            if state not in by_state:
                by_state[state] = []
            by_state[state].append(rendered)

        return {
            "tasks": [self.render(t) for t in tasks],
            "by_state": by_state,
            "totals": {
                "total": len(tasks),
                **{state: len(tasks) for state, tasks in by_state.items()},
            },
        }

    def format_for_cli(
        self,
        metrics: QueueMetrics | WorkerMetrics | TaskMetrics,
    ) -> str:
        """Format task metrics for CLI display."""
        if not isinstance(metrics, TaskMetrics):
            return "Error: Expected TaskMetrics"

        state_icon = self._status_emoji(metrics.state)

        lines = [
            f"Task: {metrics.task_id}",
            f"├── Checkpoint: {metrics.checkpoint_name}",
            f"├── State: {metrics.state} {state_icon}",
            f"├── Queue: {metrics.queue_name}",
        ]

        if metrics.worker_id:
            lines.append(f"├── Worker: {metrics.worker_id}")

        lines.append(f"├── Submitted: {self._format_time(metrics.submitted_at)}")

        if metrics.started_at:
            lines.append(f"├── Started: {self._format_time(metrics.started_at)}")
            if metrics.wait_time_ms:
                lines.append(f"├── Wait Time: {self._format_duration(metrics.wait_time_ms)}")

        if metrics.completed_at:
            lines.append(f"├── Completed: {self._format_time(metrics.completed_at)}")
            if metrics.execution_time_ms:
                lines.append(f"├── Exec Time: {self._format_duration(metrics.execution_time_ms)}")
        elif metrics.started_at:
            running_for = (datetime.now() - metrics.started_at).total_seconds() * 1000
            lines.append(f"├── Running For: {self._format_duration(running_for)}")

        if metrics.retries > 0:
            lines.append(f"├── Retries: {metrics.retries}")

        if metrics.error:
            lines.append(f"└── Error: {metrics.error[:50]}...")
        else:
            # Remove the last ├── and replace with └──
            if lines:
                lines[-1] = lines[-1].replace("├──", "└──")

        return "\n".join(lines)

    def _format_time(self, dt: datetime) -> str:
        """Format datetime for display."""
        return dt.strftime("%H:%M:%S")

    def format_task_list_for_cli(
        self,
        tasks: list[TaskMetrics],
        show_completed: bool = False,
    ) -> str:
        """Format task list for CLI display."""
        if not tasks:
            return "No tasks found."

        # Filter out completed tasks unless requested
        if not show_completed:
            tasks = [t for t in tasks if t.state not in ("succeeded", "failed", "cancelled")]

        if not tasks:
            return "No active tasks."

        lines = ["Active Tasks", "=" * 70]

        # Group by state
        by_state: dict[str, list[TaskMetrics]] = {}
        for task in tasks:
            if task.state not in by_state:
                by_state[task.state] = []
            by_state[task.state].append(task)

        for state, state_tasks in by_state.items():
            icon = self._status_emoji(state)
            lines.append(f"\n{icon} {state.upper()} ({len(state_tasks)})")
            lines.append("-" * 40)

            for task in state_tasks[:10]:  # Limit to 10 per state
                age = self._format_age(task)
                line = (
                    f"  {task.task_id[:16]:16s} "
                    f"{task.checkpoint_name[:20]:20s} "
                    f"{task.queue_name:10s} "
                    f"{age:8s}"
                )
                lines.append(line)

            if len(state_tasks) > 10:
                lines.append(f"  ... and {len(state_tasks) - 10} more")

        return "\n".join(lines)

    def _format_age(self, task: TaskMetrics) -> str:
        """Format task age."""
        if task.completed_at:
            return "done"

        if task.started_at:
            elapsed = (datetime.now() - task.started_at).total_seconds()
            return self._format_duration(elapsed * 1000)

        elapsed = (datetime.now() - task.submitted_at).total_seconds()
        return f"wait {self._format_duration(elapsed * 1000)}"

    def format_for_table(
        self,
        tasks: list[TaskMetrics],
    ) -> list[list[str]]:
        """Format task list as table rows.

        Returns:
            List of rows, first row is headers.
        """
        headers = [
            "Task ID", "Checkpoint", "State", "Queue",
            "Worker", "Wait", "Exec", "Retries"
        ]

        rows = [headers]
        for task in tasks:
            row = [
                task.task_id[:16],
                task.checkpoint_name[:20],
                task.state,
                task.queue_name,
                task.worker_id or "-",
                self._format_duration(task.wait_time_ms or 0),
                self._format_duration(task.execution_time_ms or 0),
                str(task.retries),
            ]
            rows.append(row)

        return rows



"""V1.0 to V1.1 Schema Migration.

Implements the migration logic for upgrading profile data from
schema version 1.0 to 1.1, and downgrading back.

Changes in v1.1:
- Added 'metadata' section with profiling info
- Renamed 'null_count' to 'missing_count' in columns
- Added 'data_quality_score' to each column
- Added 'semantic_type' field for ML-inferred types
- Restructured 'patterns_found' to include confidence
- Added 'recommendations' section
"""

from __future__ import annotations

import logging
from datetime import datetime
from typing import Any, Dict, List, Optional

from truthound.profiler.migration.base import (
    Migration,
    MigrationDirection,
    MigrationError,
    MigrationResult,
    SchemaVersion,
    migration_registry,
)


logger = logging.getLogger(__name__)


class V1_0_to_V1_1_Migration(Migration):
    """Migration from schema v1.0 to v1.1.

    This migration handles the following changes:

    Additions:
    - metadata: Object containing profiling metadata
    - columns[].data_quality_score: Per-column quality score
    - columns[].semantic_type: ML-inferred semantic type
    - recommendations: List of improvement recommendations

    Renames:
    - columns[].null_count -> columns[].missing_count

    Restructures:
    - patterns_found: Now includes confidence scores

    Example v1.0:
        {
            "schema_version": "1.0",
            "name": "dataset",
            "columns": [
                {"name": "id", "null_count": 0, "type": "integer"}
            ],
            "patterns_found": ["email", "phone"]
        }

    Example v1.1:
        {
            "schema_version": "1.1",
            "metadata": {
                "profiler_version": "0.2.0",
                "created_at": "2024-01-01T00:00:00",
                "migrated_from": "1.0"
            },
            "name": "dataset",
            "columns": [
                {
                    "name": "id",
                    "missing_count": 0,
                    "type": "integer",
                    "semantic_type": null,
                    "data_quality_score": 1.0
                }
            ],
            "patterns_found": [
                {"pattern": "email", "confidence": 1.0},
                {"pattern": "phone", "confidence": 1.0}
            ],
            "recommendations": []
        }
    """

    from_version = SchemaVersion(1, 0)
    to_version = SchemaVersion(1, 1)
    description = "Add metadata, semantic types, quality scores, and recommendations"
    reversible = True

    # Default values for new fields
    DEFAULT_QUALITY_SCORE = 1.0
    DEFAULT_CONFIDENCE = 1.0

    def upgrade(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Upgrade from v1.0 to v1.1.

        Args:
            data: Profile data in v1.0 format

        Returns:
            Profile data in v1.1 format
        """
        result = data.copy()

        # 1. Add metadata section
        result["metadata"] = self._create_metadata(data)

        # 2. Migrate columns
        if "columns" in result:
            result["columns"] = [
                self._upgrade_column(col) for col in result["columns"]
            ]

        # 3. Restructure patterns_found
        if "patterns_found" in result:
            result["patterns_found"] = self._upgrade_patterns(result["patterns_found"])

        # 4. Add recommendations section if missing
        if "recommendations" not in result:
            result["recommendations"] = []

        # 5. Update schema version
        result["schema_version"] = str(self.to_version)

        logger.debug(f"Upgraded profile from v1.0 to v1.1")
        return result

    def downgrade(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Downgrade from v1.1 to v1.0.

        Args:
            data: Profile data in v1.1 format

        Returns:
            Profile data in v1.0 format
        """
        result = data.copy()

        # 1. Remove metadata section
        result.pop("metadata", None)

        # 2. Migrate columns back
        if "columns" in result:
            result["columns"] = [
                self._downgrade_column(col) for col in result["columns"]
            ]

        # 3. Revert patterns_found structure
        if "patterns_found" in result:
            result["patterns_found"] = self._downgrade_patterns(result["patterns_found"])

        # 4. Remove recommendations
        result.pop("recommendations", None)

        # 5. Update schema version
        result["schema_version"] = str(self.from_version)

        logger.debug(f"Downgraded profile from v1.1 to v1.0")
        return result

    def validate_upgrade(self, data: Dict[str, Any]) -> List[str]:
        """Validate data after upgrade to v1.1.

        Args:
            data: Upgraded data

        Returns:
            List of validation errors
        """
        errors = []

        # Check metadata
        if "metadata" not in data:
            errors.append("Missing 'metadata' section")
        else:
            metadata = data["metadata"]
            if "profiler_version" not in metadata:
                errors.append("Missing 'profiler_version' in metadata")
            if "created_at" not in metadata:
                errors.append("Missing 'created_at' in metadata")

        # Check columns
        if "columns" in data:
            for i, col in enumerate(data["columns"]):
                if "missing_count" not in col:
                    errors.append(f"Column {i}: missing 'missing_count'")
                if "null_count" in col:
                    errors.append(f"Column {i}: 'null_count' should be renamed to 'missing_count'")

        # Check patterns format
        if "patterns_found" in data:
            for i, pattern in enumerate(data["patterns_found"]):
                if isinstance(pattern, str):
                    errors.append(f"Pattern {i}: should be object with 'pattern' and 'confidence'")

        return errors

    def validate_downgrade(self, data: Dict[str, Any]) -> List[str]:
        """Validate data after downgrade to v1.0.

        Args:
            data: Downgraded data

        Returns:
            List of validation errors
        """
        errors = []

        # Check no v1.1 fields remain
        if "metadata" in data:
            errors.append("'metadata' should not exist in v1.0")

        if "recommendations" in data:
            errors.append("'recommendations' should not exist in v1.0")

        # Check columns
        if "columns" in data:
            for i, col in enumerate(data["columns"]):
                if "missing_count" in col:
                    errors.append(f"Column {i}: 'missing_count' should be 'null_count' in v1.0")
                if "semantic_type" in col:
                    errors.append(f"Column {i}: 'semantic_type' should not exist in v1.0")
                if "data_quality_score" in col:
                    errors.append(f"Column {i}: 'data_quality_score' should not exist in v1.0")

        return errors

    # =========================================================================
    # Helper Methods
    # =========================================================================

    def _create_metadata(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Create metadata section for v1.1.

        Args:
            data: Original data

        Returns:
            Metadata dictionary
        """
        return {
            "profiler_version": "0.2.0",
            "created_at": datetime.now().isoformat(),
            "migrated_from": str(self.from_version),
            "original_name": data.get("name", "unknown"),
        }

    def _upgrade_column(self, column: Dict[str, Any]) -> Dict[str, Any]:
        """Upgrade a column definition from v1.0 to v1.1.

        Args:
            column: Column in v1.0 format

        Returns:
            Column in v1.1 format
        """
        result = column.copy()

        # Rename null_count to missing_count
        if "null_count" in result:
            result["missing_count"] = result.pop("null_count")

        # Add data_quality_score if missing
        if "data_quality_score" not in result:
            # Calculate from existing data if possible
            result["data_quality_score"] = self._calculate_column_quality(result)

        # Add semantic_type if missing
        if "semantic_type" not in result:
            result["semantic_type"] = self._infer_semantic_type(result)

        return result

    def _downgrade_column(self, column: Dict[str, Any]) -> Dict[str, Any]:
        """Downgrade a column definition from v1.1 to v1.0.

        Args:
            column: Column in v1.1 format

        Returns:
            Column in v1.0 format
        """
        result = column.copy()

        # Rename missing_count back to null_count
        if "missing_count" in result:
            result["null_count"] = result.pop("missing_count")

        # Remove v1.1 fields
        result.pop("data_quality_score", None)
        result.pop("semantic_type", None)

        return result

    def _upgrade_patterns(
        self,
        patterns: List[Any],
    ) -> List[Dict[str, Any]]:
        """Upgrade patterns_found from v1.0 to v1.1 format.

        Args:
            patterns: List of patterns (strings in v1.0)

        Returns:
            List of pattern objects with confidence
        """
        result = []

        for pattern in patterns:
            if isinstance(pattern, str):
                # Convert string to object
                result.append({
                    "pattern": pattern,
                    "confidence": self.DEFAULT_CONFIDENCE,
                    "migrated": True,
                })
            elif isinstance(pattern, dict):
                # Already in new format
                if "confidence" not in pattern:
                    pattern["confidence"] = self.DEFAULT_CONFIDENCE
                result.append(pattern)
            else:
                # Unknown format, try to convert
                result.append({
                    "pattern": str(pattern),
                    "confidence": self.DEFAULT_CONFIDENCE,
                    "migrated": True,
                })

        return result

    def _downgrade_patterns(
        self,
        patterns: List[Any],
    ) -> List[str]:
        """Downgrade patterns_found from v1.1 to v1.0 format.

        Args:
            patterns: List of pattern objects

        Returns:
            List of pattern strings
        """
        result = []

        for pattern in patterns:
            if isinstance(pattern, dict):
                result.append(pattern.get("pattern", str(pattern)))
            else:
                result.append(str(pattern))

        return result

    def _calculate_column_quality(self, column: Dict[str, Any]) -> float:
        """Calculate quality score for a column.

        Args:
            column: Column data

        Returns:
            Quality score (0.0 to 1.0)
        """
        scores = []

        # Completeness (from null/missing ratio)
        missing_count = column.get("missing_count", column.get("null_count", 0))
        row_count = column.get("row_count", 0)

        if row_count > 0:
            completeness = 1.0 - (missing_count / row_count)
            scores.append(completeness)
        else:
            scores.append(self.DEFAULT_QUALITY_SCORE)

        # Uniqueness (if available)
        if "unique_ratio" in column:
            # Higher uniqueness is generally better for IDs, not for categories
            scores.append(min(1.0, column["unique_ratio"] * 1.2))

        # Return average
        return sum(scores) / len(scores) if scores else self.DEFAULT_QUALITY_SCORE

    def _infer_semantic_type(self, column: Dict[str, Any]) -> Optional[str]:
        """Infer semantic type from column metadata.

        This provides a simple inference based on column name and type.
        More sophisticated inference would use the ML module.

        Args:
            column: Column data

        Returns:
            Inferred semantic type or None
        """
        name = column.get("name", "").lower()
        dtype = column.get("type", column.get("physical_type", "")).lower()

        # Name-based inference
        if "email" in name:
            return "email"
        if "phone" in name or "tel" in name:
            return "phone"
        if "url" in name or "link" in name or "website" in name:
            return "url"
        if "uuid" in name or "guid" in name:
            return "uuid"
        if "id" in name or name.endswith("_id"):
            return "identifier"
        if "date" in name or "time" in name:
            return "datetime"
        if "price" in name or "cost" in name or "amount" in name:
            return "currency"
        if "percent" in name or "ratio" in name or "rate" in name:
            return "percentage"
        if "status" in name or "type" in name or "category" in name:
            return "categorical"

        # Type-based inference
        if "int" in dtype:
            return "integer"
        if "float" in dtype or "double" in dtype:
            return "float"
        if "bool" in dtype:
            return "boolean"
        if "date" in dtype:
            return "datetime"

        return None


# =============================================================================
# Additional Migrations (Future)
# =============================================================================


class V1_1_to_V1_2_Migration(Migration):
    """Placeholder for future v1.1 to v1.2 migration.

    This demonstrates the pattern for adding new migrations.
    """

    from_version = SchemaVersion(1, 1)
    to_version = SchemaVersion(1, 2)
    description = "Future migration placeholder"
    reversible = True

    def upgrade(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Upgrade from v1.1 to v1.2."""
        result = data.copy()
        result["schema_version"] = str(self.to_version)
        # Future changes would go here
        return result

    def downgrade(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """Downgrade from v1.2 to v1.1."""
        result = data.copy()
        result["schema_version"] = str(self.from_version)
        # Reverse future changes here
        return result

    def validate_upgrade(self, data: Dict[str, Any]) -> List[str]:
        """Validate upgrade to v1.2."""
        return []

    def validate_downgrade(self, data: Dict[str, Any]) -> List[str]:
        """Validate downgrade to v1.1."""
        return []


# =============================================================================
# Register Migrations
# =============================================================================


def register_migrations() -> None:
    """Register all migrations with the global registry."""
    migration_registry.register(V1_0_to_V1_1_Migration())
    migration_registry.register(V1_1_to_V1_2_Migration())


# Auto-register on import
register_migrations()
